import pandas as pd
from numpy import array, max
import pytest
import sys
import os
sys.path.append(".")
from trendfitter.models import PCA

TESTDATA1_FILENAME = os.path.join(os.path.dirname(__file__), 'pca_test_dataset_complete.csv')
TESTDATA2_FILENAME = os.path.join(os.path.dirname(__file__), 'pca_test_dataset_incomplete.csv')

class TestPCA(object):

    def test_fit_fulldataset(self):
        
        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col=0)
        test_data = (test_data - test_data.mean()) / test_data.std()
        test_model = PCA()
        test_model.fit(test_data)

        assert test_model.principal_components == 2, 'Number of principal components is not as expected'

        first_2_comps = array([[-0.45753346,  0.47874552, -0.53238765,  0.50447686, -0.15340256], 
                               [-0.37043881,  0.35674995,  0.19766106, -0.22123996,  0.80466611]])
        assert test_model.loadings[:2, :] == pytest.approx(first_2_comps), 'Loadings are not as expected'

        expected_VIPs = array([[0.52334216, 0.57299319, 0.70859153, 0.63624226, 0.05883086]])
        assert test_model.feature_importances_ == pytest.approx(expected_VIPs), 'VIPs are not as expected'

        expected_omega = array([[ 1.48529445e+02, -6.14879643e-06],
                                [-6.14879643e-06,  6.34895824e+01]])

        #assert test_model.omega == pytest.approx(expected_omega), 'Omega matrix is not as expected'

        expected_q2 = array([0.5567342018714448, 0.845616530550494])
        assert test_model.q2 == pytest.approx(expected_q2), 'Q2 results are not the expected'

        expected_chi2 = array([[1.9294110960533677, 3.1291520687831618], 
                               [0.6596194475295534, 0.3878793546907029]])
        assert test_model._chi2_params == pytest.approx(expected_chi2), 'Chi2 parameters are not as expected'   
    
    def test_fit_missingdataset(self):

        test_data = pd.read_csv(TESTDATA2_FILENAME, index_col = 0, delimiter = ';')
        test_data = (test_data - test_data.mean()) / test_data.std()
        test_model = PCA()
        test_model.fit(test_data)

        assert test_model.principal_components == 2, 'Number of principal components is not as expected'

        first_2_comps = array([[-0.4632267,   0.48424391, -0.53101657,  0.50288033, -0.12673451],
                               [-0.36939071,  0.32991466,  0.20386669, -0.25984463,  0.80350853]])
        assert test_model.loadings[:2, :] == pytest.approx(first_2_comps), 'Loadings are not as expected'

        expected_VIPs = array([[0.53644745, 0.5862304,  0.7049465,  0.63222156, 0.04015409]])
        assert test_model.feature_importances_ == pytest.approx(expected_VIPs), 'VIPs are not as expected'

        expected_omega = array([[ 1.47869724e+02, -4.00782741e+00],
                                [-4.00782741e+00,  6.45928321e+01]])
        assert test_model.omega == pytest.approx(expected_omega), 'Omega matrix is not as expected'

        expected_q2 = array([0.5572977189894531, 0.8418633512116621])
        assert test_model.q2 == pytest.approx(expected_q2), 'Q2 results are not the expected'

        expected_chi2 = array([[1.8519373459166755, 2.9781940475342488], 
                               [0.7473545122266293, 0.5696801312124594]])
        assert test_model._chi2_params == pytest.approx(expected_chi2), 'Chi2 parameters are not as expected'   

    
    def test_predict_fulldata(self):
        
        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col=0)
        test_data = (test_data - test_data.mean()) / test_data.std()
        test_model = PCA()
        test_model.fit(test_data)

        expected = array([[-0.40341328,  0.44123301, -0.858837,    0.8348354,  -0.71060328],
                          [ 1.14753328, -1.211652,    1.55766262, -1.488007,    0.71331028],
                          [-0.21042758,  0.19359614,  0.29674882, -0.31043066,  0.72962799],
                          [-0.46436954,  0.52113604, -1.25815783,  1.2309502,  -1.21621669],
                          [-0.82589664,  0.84371014, -0.54389296,  0.49286008,  0.3393626 ]])

        assert expected == pytest.approx(test_model.predict(test_data.iloc[:5])), 'Prediction of X values unsuccessful'
    
    
    def test_predict_missingdata(self):
        
        test_data = pd.read_csv(TESTDATA2_FILENAME, index_col = 0, delimiter = ';')
        test_data = (test_data - test_data.mean()) / test_data.std()
        test_model = PCA()
        test_model.fit(test_data)

        expected = array([[-0.6253456,   0.6455792,  -0.6260675,   0.5832297,  -0.04016798],
                          [ 1.1948117,  -1.26079601,  1.50100527, -1.43545535,  0.5162801 ],
                          [-0.11255508,  0.08932129,  0.1871146,  -0.21085459,  0.42507269],
                          [-0.68438244, 0.73650655, -1.01960558,  0.99060505, -0.52620305],
                          [-0.70410329,  0.72388327, -0.67142978,  0.62140642,  0.003059486]])

        assert expected == pytest.approx(test_model.predict(test_data.iloc[:5])), 'Prediction of X values with missing data unsuccessful'
    
    
    def test_transform_fulldata(self):

        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col=0)
        test_data = (test_data - test_data.mean()) / test_data.std()
        test_model = PCA()
        test_model.fit(test_data)
        
        expected = array([[ 1.38321112, -0.61940618],
                          [-2.79447691,  0.35372482],
                          [-0.23755635,  0.86145823],
                          [ 1.93934048, -1.14173678],
                          [ 1.26793711,  0.66346449]])

        assert expected == pytest.approx(test_model.transform(test_data.iloc[:5]))

    def test_transform_missingdata(self):

        test_data = pd.read_csv(TESTDATA2_FILENAME, index_col = 0, delimiter = ';')
        test_data = (test_data - test_data.mean()) / test_data.std()
        test_model = PCA()
        test_model.fit(test_data)

        expected = array([[ 1.23456343,  0.14473251],
                          [-2.74628248,  0.20937094],
                          [-0.1588916,   0.50395936],
                          [ 1.77623882, -0.37472197],
                          [ 1.34748053,  0.21634091]])

        assert expected == pytest.approx(test_model.transform(test_data.iloc[:5]))

    
    def test_score(self):
        
        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col=0)
        test_data = (test_data - test_data.mean()) / test_data.std()
        test_model = PCA()
        test_model.fit(test_data)
        
        expected = 0.8527862417379573
        
        assert expected == pytest.approx(test_model.score(test_data.iloc[:5]))
        
    
    def test_Hot_T2(self):
        
        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col=0)
        test_data = (test_data - test_data.mean()) / test_data.std()
        test_model = PCA()
        test_model.fit(test_data)

        expected = array([0.94621912, 2.72734253, 0.60343192, 2.29268863, 0.88785326])
        expected_95_lim = 6.627785016879078

        assert expected == pytest.approx(test_model.Hotellings_T2(test_data.iloc[:5]))
        assert expected_95_lim == pytest.approx(test_model.T2_limit(0.95))

    def test_SPEs(self):

        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col=0)
        test_data = (test_data - test_data.mean()) / test_data.std()
        test_model = PCA()
        test_model.fit(test_data)

        expected = array([0.39903279, 1.46216746, 0.40595311, 0.5964136,  0.35905571])
        expected_95_lim = 2.868556708846775

        assert expected == pytest.approx(test_model.SPEs(test_data.iloc[:5]))
        assert expected_95_lim == pytest.approx(test_model.SPE_limit(0.99))
    
    def test_contributions_scores_ind(self):
        
        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col=0)
        test_data = (test_data - test_data.mean()) / test_data.std()
        test_model = PCA()
        test_model.fit(test_data)
        
        expected = array([[-0.18514252,  0.33746791, -0.37740461,  0.16550264, -0.45989523],
                          [ 0.23488624, -1.24472941,  1.20875952, -1.7811052,   0.12322141],
                          [-0.18260662,  0.02794715,  0.04541179, -0.12913747,  0.29305082],
                          [-0.2008982,   0.32514272, -0.54072132,  1.13215426, -0.88831667],
                          [-0.22724196,  0.38668715, -0.11965834,  0.36985629,  0.23187505]])

        assert expected == pytest.approx(test_model.contributions_scores_ind(test_data.iloc[:5]))
    
    def test_contributions_spe(self):

        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col=0)
        test_data = (test_data - test_data.mean()) / test_data.std()
        test_model = PCA()
        test_model.fit(test_data)

        expected = array([[-1.40922213e-03,  1.16346779e-01,  7.69758438e-06, -1.96541047e-01,  -8.47280388e-02],
                          [-6.96758397e-01, -1.40997887e-01, -2.58981400e-02, -4.64770291e-01,  -1.33742743e-01],
                          [-1.75533104e-01, -8.83557269e-03, -6.97661053e-04, -1.56612937e-01,  -6.42738372e-02],
                          [ 2.22141917e-02, -3.97250224e-04,  1.61680615e-01,  3.89537193e-01,   2.25843449e-02],
                          [ 6.72448679e-02,  9.85292592e-03,  6.30157900e-02,  2.00243599e-01,   1.86985234e-02]])

        assert expected == pytest.approx(test_model.contributions_spe(test_data.iloc[:5]))


from numpy import array, isnan, nansum, nan_to_num, multiply, sum, sqrt, append, zeros, place, nan, concatenate, mean, nanvar, std, unique, where, nanmean, identity
from numpy.linalg import norm, pinv
from sklearn.model_selection import KFold
from pandas import DataFrame, Series
from trendfitter.auxiliary.tf_aux import scores_with_missing_values

class SMB_PLS:
    """
    A sklearn-like class for the Sequential Multi-Block Projection to Latent Structures.

        Parameters
        ----------
        cv_splits_number : int, optional
            number of splits used for cross validation in case latent_variables is None
        tol : float, optional
            value used to decide if model has converged
        loop_limit : int, optional
            maximum number of loops for the extraction of one latent variable
        missing_values_method : str, optional
            string to define the method that the model will deal with missing values

        Attributes
        ----------
        latent_variables : [int]
            list of number of latent variables deemed relevant from each block. 
        block_divs : [int]
            list with the index of every block final position. ex. [2,4] means two blocks with 
                columns 0 and 1 on the first block and columns 2 and 3 on the second block. 
                Assigned when fit method runs.
        p_loadings_block : [array_like]
            list of p loadings arrays of every block with all the extracted latent variables 
        superlevel_p_loadings : array_like
            array of all latent variables extracted p loadings for the super level
        weights_block : [array_like]
            list of weights arrays of every block with all the extracted latent variables 
        weights_super : array_like
            array of all latent variables extracted weights for the super level
        c_loadings : array_like
            array of c_loadings for all extracted latent variables
        q2 : [float]
            list of r² coefficients extracted during cross validation
        feature_importances : [float]
            list of values that represent how important is each variable in the same order 
                of the X columns on the first matrix

        Methods
        -------
        fit(X, blocks_sep, Y)
            Applies the NIPALS like method to find the best parameters that adjust the model 
                to the data
        transform(X)
            Transforms the original data to the latent variable space in the super level 
        transform_inv(scores)
            Returns the superlevel scores to the original data space
        transform_b(X, block)
            Transforms the original data to the latent variable space in the block level for
                all blocks
        predict(X)
            Predicts Y values 
        score(X, Y)
            calculates the r² value for Y
        Hotellings_T2(X)
            Calculates Hotellings_T2 values for the X data in the super level
        Hotellings_T2_blocks(X)
            Calculates Hotellings_T2 values for in the block level for all blocks
        SPEs_X(X)
            Calculates squared prediction errors on the X side for the super level
        SPEs_X_blocks(X)
            Calculates squared prediction errors for the block level for all blocks
        SPEs_Y(X, Y)
            Calculates squared prediction errors for the predictions
        contributions_scores(X)
            calculates the contributions of each variable to the scores on the super level
        contributions_scores_b(X)
            calculates the contributions of each variable to the scores on the super level
        contributions_SPE(X)
            calculates the contributions of each variable to the SPE on the X side for
                the super level
        contributions_SPE_b(X)
            calculates the contributions of each variable to the SPE on the X side for
                the block level for all blocks

    """
    def __init__(self, cv_splits_number = 7, tol = 1e-16, loop_limit = 1000, missing_values_method = 'TSM'):
        
        # Parameters
        
        self.cv_splits_number = cv_splits_number # number of splits for latent variable cross-validation
        self.tol = tol # criteria for convergence
        self.loop_limit = loop_limit # maximum number of loops before convergence is decided to be not attainable
        self.q2y = [] # list of cross validation scores
        self.missing_values_method = missing_values_method
        
        # Attributes

        self.latent_variables = None # number of principal components to be extracted
        self.block_divs = None
        self.block_p_loadings = None
        self.superlevel_p_loadings = None
        self.block_weights = None
        self.block_weights_star = None
        self.superlevel_weights = None
        self.x_weights_star = None
        self.x_weights = None
        self.c_loadings = None 
        self.VIPs = None
        self.coefficients = None
        self.omega = None # score covariance matrix for missing value estimation
        self.training_superlevel_scores = None
        self.training_y_scores = None
            
    def fit(self, X, block_divs, Y, latent_variables = None, deflation = 'both', int_call = False, random_state = None):
        """
        Adjusts the model parameters to best fit the Y using the algorithm defined in 
            Lauzon-Gauthier et al's [1]

        Parameters
        ----------
        X : array_like
            Matrix with all the data to be used as predictors in one only object
        block_divs : [int]
            list with the index of every block final position. ex. [2,4] means two blocks with 
                columns 0 and 1 on the first block and columns 2 and 3 on the second block.
        Y : array_like
            Matrix with all the data to be predicted in one only object
        latent_variables : [int], optional
            list of number of latent variables deemed relevant from each block. If left unspecified
                a cross validation routine will define the number during fitting
        deflation : str, optional
            string defining method of deflation, only Y or both X and Y 
        int_call : Boolean, optional
            Flag to define if it is an internal call on the cross validation routine and decide
                if it is necessary to calculate the VIP values

        References 
        ----------
        [1] J. Lauzon-Gauthier, P. Manolescu, C. Duchesne, "The Sequential Multi-block PLS 
            algorithm (SMB-PLS): Comparison of performance and interpretability," Chemometrics 
            and Intelligent Laboratory Systems, vol 180, pp. 72-83, 2018.

        """

        
        # check if X and Y make sense
        # check if block_divs make sense 
        # check if latent_variables make sense


        self.block_divs = block_divs
        block_coord_pairs = (*zip([0] + block_divs[:-1], block_divs),)
        G = [identity(pair[1] - pair[0]) for pair in block_coord_pairs]

        if isinstance(X, DataFrame):# If X data is a pandas Dataframe
            indexes = X.index.copy()
            X_columns = X.columns.copy()
            X = array(X.to_numpy(), ndmin = 2)

        missing_values_list = [isnan(sum(X[:, start:end])) for (start, end) in block_coord_pairs] 
          
        if isinstance(Y, DataFrame) or isinstance(Y, Series): # If Y data is a pandas Dataframe or Series 
            if isinstance(Y, DataFrame) : Y_columns = Y.columns
            Y = array(Y.to_numpy(), ndmin = 2).T
            
        Orig_X, X = X.copy(), X.copy()
        Orig_Y, Y = Y.copy(), Y.copy() 

        missing_values_list = missing_values_list + [isnan(sum(Y))] # This is a check for missing data it'll allow for dealing with it in the next steps
        
        """------------------- Handling the case where the amount of latent variables is not defined -------------------"""
        if latent_variables != None : 
            self.latent_variables = latent_variables 
            max_LVs = latent_variables
        else:
            max_LVs = array(block_divs) - array([0] + block_divs[:-1])
            
            
        """------------------- model Calculation -------------"""
        q2_final = []
        LV_valid = [0 for _ in block_divs]
        for block, LVs in enumerate(max_LVs):

            for LV in range(LVs):

                if not self.latent_variables is None and  LV > 2 * self.latent_variables[block]: break

                result = self._SMBPLS_1LV(X, block_coord_pairs[block:], Y, missing_values_list, block)

                #--------------------------cross-validation------------------------------
                if (block != 0 or LV != 0) and self.latent_variables is None and not int_call:

                    LV_valid[block] = LV + 1
                    q2_final.append(self._cross_validate(Orig_X, block_divs, Orig_Y, LV_valid, random_state))

                    if (q2_final[-1] < q2_final[-2] or  # mean cross-validation performance of the model has decreased with the addition of another latent variable
                        q2_final[-1] - q2_final[-2] < 0.01 or  # mean cross-validation performance of the model has increased less than 1% with the addition of another latent variable
                        LV > (block_coord_pairs[block][1] - block_coord_pairs[block][0]) // 2): # the amount of latent variables added is more than half the variables on X
                        
                        q2_final = q2_final[:-1]
                        self.q2y = q2_final
                        LV_valid[block] -= 1
                        if block == len(max_LVs) - 1 : self.latent_variables = LV_valid 
                        if self.missing_values_method == 'TSM': break  #In case of TSM use there is no need of more components for missing value estimation
                elif int_call and not (latent_variables is None):
                    LV_valid[block] = LV + 1
                    #q2_final.append(self._cross_validate(Orig_X, block_divs, Orig_Y, LV_valid, random_state))
                else:
                    LV_valid[block] = LV + 1    
                    q2_final.append(self._cross_validate(Orig_X, block_divs, Orig_Y, LV_valid, random_state))
                    
                #------------------------------------deflation----------------------------

                if deflation == 'both' :
                    X -= result['tT'] @ result['p']
                    Y -= result['tT'] @ result['c']
                elif deflation == 'Y':
                    Y -= result['tT'] @ result['c']
                else : raise Exception(f'Deflation method "{deflation}" non-existing.')
                    
                #--------------------------Property assignment section--------------------
                if LV == 0 and block == 0 :
                    self.superlevel_weights = result['wT']
                    self.x_weights = result['w']
                    self.block_weights = result['wb']
                    self.block_p_loadings = result['pb']
                    self.superlevel_p_loadings = result['p']
                    self.c_loadings = result['c']
                    self.training_superlevel_scores = result['tT']
                    self.training_block_scores = result['tb']
                    self.training_y_scores = result['u']
                    self.x_weights_star2 = result['w']
                    for bl, (start, end) in enumerate(block_coord_pairs):
                        G[bl] = G[bl] - result['w'][:, start:end].T @ result['p'][:, start:end]

                
                else:                    
                    self.superlevel_weights = append(self.superlevel_weights, result['wT'], axis = 0)
                    self.x_weights = append(self.x_weights, result['w'], axis = 0)
                    self.block_weights = append(self.block_weights, result['wb'], axis = 0)
                    self.block_p_loadings = append(self.block_p_loadings, result['pb'], axis = 0)
                    self.superlevel_p_loadings = append(self.superlevel_p_loadings, result['p'], axis = 0)
                    self.c_loadings = append(self.c_loadings, result['c'], axis = 0)
                    self.training_superlevel_scores = append(self.training_superlevel_scores, result['tT'], axis = 1)
                    self.training_block_scores = append(self.training_block_scores, result['tb'], axis = 1)
                    self.training_y_scores = append(self.training_y_scores, result['u'], axis = 1)

                    for bl, (start, end) in enumerate(block_coord_pairs[block:]):
                        if bl == 0: x_weights_star2 = array(G[block + bl] @ result['w'][:, start:end].T, ndmin = 2).T
                        else: x_weights_star2 = append(x_weights_star2, array(G[block + bl] @ result['w'][:, start:end].T, ndmin = 2 ).T, axis = 1)
                        if block > 0 and bl == 0: x_weights_star2 = append(zeros((1, start)), x_weights_star2, axis = 1)

                        G[block + bl] = G[block + bl] - result['w'][:, start:end].T @ result['p'][:, start:end]
                    
                    self.x_weights_star2 = append(self.x_weights_star2, x_weights_star2, axis = 0)
                    


        
        self.block_weights_star = (self.block_weights @ pinv(self.block_p_loadings.T @ self.block_weights))
        self.x_weights_star = (self.x_weights @ pinv(self.superlevel_p_loadings.T @ self.x_weights))

        return

    def transform(self, X, latent_variables = None): 
        
        """
        Transforms the X matrix to the model-fitted space returning scores
        of the super level.

        Parameters
        ----------
        X : array_like
            Matrix with all the data to be used as predictors in one only object
        latent_variables : [int], optional
            list with number of latent variables to be used. 

        Returns
        -------
        result : array_like of shape (X.shape[0], sum(latent_variables))
            Scores for the X values transformed on the super level

        """

        if isinstance(X, DataFrame): 
            X_values = X.to_numpy()
        else:
            X_values = X    

        if latent_variables is None : 
            latent_variables = sum(self.latent_variables)
        else:
            latent_variables = sum(latent_variables)
        
        if isnan(sum(X_values)):
            
            result = zeros((X_values.shape[0], latent_variables))
            X_nan = isnan(X_values)
            variables_missing_mask = unique(X_nan, axis = 0)

            for row_mask in variables_missing_mask:
                
                rows_indexes = where((X_nan == row_mask).all(axis = 1))                
                
                if sum(row_mask) == 0: 

                    result[rows_indexes, :] = X[rows_indexes, :] @ self.x_weights_star[:latent_variables, :].T 
                
                else:
                    
                    result[rows_indexes, :] = scores_with_missing_values(self.omega, self.x_weights_star[:, ~row_mask], X[rows_indexes[0][:, None], ~row_mask], 
                                                                            LVs = latent_variables, method = self.missing_values_method)
                    
        else : result = X_values @ self.x_weights_star[:latent_variables, :].T

        # TO DO : check if X makes sense with latent variables
        return result

    def transform_inv(self, scores, latent_variables = None):

        """
        Transforms the superlevel scores matrix to the original X.

        Parameters
        ----------
        scores : array_like
            Matrix with all the scores to be used to rebuild X
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        result : array_like 
            matrix of rebuilt X from scores
        """
                  
        if latent_variables is None : 
            latent_variables = sum(self.latent_variables)
        else:
            latent_variables = sum(latent_variables)

        result = scores @ self.p[:latent_variables, :] 
        
        return result
    
    def transform_b( self, X, latent_variables = None ): # To Do

        pass
                    
    def predict( self, X, latent_variables = None): 

        """
        Predicts Y values using X array.

        Parameters
        ----------
        X : array_like
            Samples Matrix
        latent_variables : [int], optional
            number of latent variables to be used. 

        Returns
        -------
        preds : array_like 
            returns predictions
        """

        if isinstance( X, DataFrame ) : X = X.to_numpy()        
        if latent_variables is None : 
            latent_variables = sum(self.latent_variables)
        else:
            latent_variables = sum(latent_variables)

        preds = self.transform(X, latent_variables = latent_variables) @ self.c_loadings[ :latent_variables, :]
        
        return preds

    def score(self, X, Y, latent_variables = None): 

        """
        Return the coefficient of determination R^2 of the prediction.

        R² is defined as 1 - Variance(Error) / Variance(Y) with Error = Y - predictions(X)

        Parameters
        ----------
        X : array_like
            Matrix with all the X to be used
        Y : array_like
            Matrix with all the Y ground truth values
        latent_variables : [int], optional
            number of latent variables to be used. 

        Returns
        -------
        score : float 
            returns calculated r².
        """

        if latent_variables is None : latent_variables = self.latent_variables
        if isinstance(Y, DataFrame) or isinstance(Y, Series): 
            Y_values = array(Y.to_numpy(), ndmin = 2).T
        else: 
            Y_values = Y

        Y_hat = self.predict(X, latent_variables = latent_variables)
        F = Y_values - Y_hat
        score = 1 - nanvar(F) / nanvar(Y_values)

        return score
    
    def Hotellings_T2(self, X, latent_variables = None): 
        
        """
        Calculates the Hotelling's T² for the X samples on the superlevel.

        Parameters
        ----------
        X : array_like
            Samples Matrix
        latent_variables : [int], optional
            number of latent variables to be used. 

        Returns
        -------
        T2s : array_like 
            returns all calculated T²s for the X samples
        """
        
        if isinstance(X, DataFrame) : X = X.to_numpy()     #dataframe, return it
   
        if latent_variables is None : latent_variables = self.latent_variables # Unless specified, the number of PCs is the one in the trained model 
        
        scores_matrix = self.transform( X, latent_variables = latent_variables)
        
        T2s = array(sum(((scores_matrix / std(scores_matrix)) ** 2), axis = 1), ndmin = 2).T
        
        return T2s
    
    #def Hotellings_T2_block(self, X, block, latent_variables = None):

        """
        Calculates the Hotelling's T² for the X samples on the superlevel.

        Parameters
        ----------
        X : array_like
            Samples Matrix
        latent_variables : [int], optional
            number of latent variables to be used. 

        Returns
        -------
        T2s : array_like 
            returns all calculated T²s for the X samples
        """


    #    return T2s_block
        
    def SPEs_X(self, X, latent_variables = None): 
        
        """
        Calculates the Squared prediction error for for the X matrix rebuild.

        Parameters
        ----------
        X : array_like
            Samples Matrix
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        SPE : array_like 
            returns all calculated SPEs for the X samples
        """
        
        if latent_variables is None : latent_variables = self.latent_variables

        if isinstance(X, DataFrame) : X = X.to_numpy()
        
        
        error = X - self.transform_inv(self.transform(X, latent_variables = latent_variables), latent_variables = latent_variables)   
        SPE = nansum(error ** 2, axis = 1)
        
        return SPE

    def SPEs_Y(self, X, Y, latent_variables = None) :

        """
        Calculates the Squared prediction error for the Y values.

        Parameters
        ----------
        X : array_like
            Samples X Matrix
        Y : array_like
            ground truth Y Matrix
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        SPE : array_like 
            returns all calculated SPEs for the X samples
        """
        
        if latent_variables is None : latent_variables = self.latent_variables

        if isinstance(X, DataFrame) : X = X.to_numpy()
        
        
        error = Y - self.predict(X, latent_variables = latent_variables)
        SPE = nansum(error ** 2, axis = 1)
        
        return SPE
 
    def contributions_scores_ind(self, X, latent_variables = None): 
        
        """
        calculates the sample individual contributions to the super level scores.

        Parameters
        ----------
        X : array_like
            Sample matrix 
        latent_variables : [int], optional
            List of number of latent variables to be used. 

        Returns
        -------
        contributions : array_like 
            matrix of scores contributions for every X sample
        """

        if latent_variables is None : latent_variables = self.latent_variables
        if isinstance(X, DataFrame) : X = X.to_numpy()

        scores = self.transform(X, latent_variables = latent_variables)
        scores = (scores / std(scores, axis = 0)) ** 2

        if latent_variables is None : 
            latent_variables = sum(self.latent_variables)
        else:
            latent_variables = sum(latent_variables)

        contributions = X * (scores @ (self.x_weights_star[:latent_variables, :] ** 2) ** (1 / 2))

        return contributions
    
    def contributions_SPE_X(self, X, latent_variables = None): 
        
        """
        calculates the individual sample individual contributions to the squared prediction error 
            of the X variables matrix reconstruction.

        Parameters
        ----------
        X : array_like
            Sample matrix 
        latent_variables : [int], optional
            number of latent variables to be used. 

        Returns
        -------
        SPE_contributions : array_like 
            matrix of SPE contributions for every X sample
        """

        if latent_variables is None : latent_variables = self.latent_variables
        if isinstance(X, DataFrame) : X = X.to_numpy()
        
        error = X - self.transform_inv(self.transform(X))
        
        SPE_contributions = (error ** 2) * where(error > 0, 1, -1)
               
        return SPE_contributions

    def _SMBPLS_1LV(self, X, block_coord_pairs, Y, missing_values, block):
        
        conv = 1
        loops = 0

        y_scores1 = nan_to_num(array(Y[ :,0 ], ndmin = 2).T) #handling possible NaNs that come from faulty datasets - Step 1
        superlevel_scores1 = nan_to_num(y_scores1) #initializing superlevel scores vector

        while (conv > self.tol and loops < self.loop_limit) :

            first_block_done = False
            for missing, (start, end) in zip(missing_values, block_coord_pairs) :

                # calculates the block weights and scores for the uncorrelated part 
                if not first_block_done: 
                    block_weights, T_scores = self._uncorrelated_part(X[:, start:end], y_scores1, missing)
                    first_block_done = True
                    continue

                # calculating the block weights and scores for the blocks following the correlated parts of the following blocks
                corr_block_weights, corr_T_scores =  self._correlated_part(array(T_scores[:, 0], ndmin = 2).T, X[:, start:end], y_scores1, missing)
                block_weights = append(block_weights, corr_block_weights, axis = 1)
                T_scores = append(T_scores, corr_T_scores, axis = 1)
            
            # calculating the superlevel weights and scores
            superlevel_weights, superlevel_scores = self._superlevel_part(T_scores, y_scores1)

            # calculating the y side loadings and scores
            c_loadings, y_scores = self._y_part(superlevel_scores, Y, missing_values[-1])

            conv = norm(superlevel_scores1 - superlevel_scores) / norm(superlevel_scores1)
            superlevel_scores1 = superlevel_scores
            y_scores1 = y_scores

        # Calculating the p_loadings that connect from the raw X values to the superlevel scores
        start, end = block_coord_pairs[0][0], block_coord_pairs[-1][-1]
        superlevel_p_loadings = self._p_loadings(X[:, start:end], superlevel_scores, True in missing_values[:-1]).T
        x_weights = self._p_loadings(X[:, start:end], y_scores, True in missing_values).T
        
        x_weights = x_weights / norm(x_weights)
        y_scores_test = X[:, start:end] @ x_weights

        for missing, scores, (start, end) in zip(missing_values, T_scores.T, block_coord_pairs) :
            
            if start == block_coord_pairs[0][0] : 
                block_p_loadings = self._p_loadings(X[:, start:end], array(scores, ndmin = 2).T, missing)
                if start > 0:
                    block_p_loadings = append(zeros((1, start)), block_p_loadings, axis = 1 )
                    superlevel_p_loadings = append(zeros((start, 1)), superlevel_p_loadings, axis = 0 )
                    x_weights = append(zeros((start, 1)), x_weights, axis = 0 )
                    superlevel_weights = append(zeros((block, 1)), superlevel_weights, axis = 0 )
                    block_weights = append(zeros((1, start)), block_weights, axis = 1 )

            else : 
                block_p_loadings = append(block_p_loadings, self._p_loadings(X[:, start:end], array(scores, ndmin = 2).T, missing), axis = 1)

        result_dict = {'wb':block_weights,
                       'w':x_weights.T,
                       'pb':block_p_loadings,
                       'tb':T_scores,
                       'wT':superlevel_weights.T,
                       'p':superlevel_p_loadings.T,
                       'tT':superlevel_scores,
                       'u':y_scores,
                       'c':c_loadings.T}

        return result_dict
    
    def _uncorrelated_part(self, X, y_scores, missing_value):

        if missing_value:
            block_weights = array(nansum((X * y_scores), axis = 0), ndmin = 2) 
            block_weights = block_weights / nansum((array(~isnan(sum(X, axis = 1)), ndmin = 2).T * y_scores) ** 2, axis = 0)
            block_weights = block_weights / norm(block_weights)
            T_scores = array(nansum(X * block_weights, axis = 1) / nansum(((~isnan(X) * block_weights) ** 2), axis = 1), ndmin = 2).T
        
        else:
            block_weights = X.T @ y_scores / (y_scores.T @ y_scores) # calculating Xb block weights (as step 2.1 in L-G's 2018 paper)
            block_weights = (block_weights / norm(block_weights)).T
            T_scores = X @ block_weights.T
        
        return block_weights, T_scores
    
    def _correlated_part(self, scores, X, y_scores, missing_value):
        
        X_corr_coeffs = (scores / (scores.T @ scores)) @ scores.T
        
        if missing_value :     
            X_corr = X_corr_coeffs @ nan_to_num(X)  # Attention on this part
            place(X_corr, isnan(X), nan) # Keeping the NaN value as an NaN
            block_weights = array(nansum((X_corr * y_scores), axis = 0), ndmin = 2) 
            block_weights = block_weights / nansum((array(~isnan(sum(X, axis = 1)), ndmin = 2).T * y_scores) ** 2, axis = 0)
            block_weights = block_weights / norm(block_weights) #step 2.6
            T_scores =  array(nansum(X_corr * block_weights.T, axis = 1), ndmin = 2).T # step 2.7
        else :    
            X_corr = X_corr_coeffs @ X # finishing step 2.4 for no missing data
            block_weights = X_corr.T @ y_scores / (y_scores.T @ y_scores) # step 2.5
            block_weights = (block_weights / norm(block_weights)).T #step 2.6
            T_scores =  X_corr @ block_weights.T # step 2.7                
        
        return block_weights, T_scores
    
    def _superlevel_part(self, T_scores, y_scores):
        
        superlevel_weights = T_scores.T @ y_scores / (y_scores.T @ y_scores)  #step 2.9
        superlevel_weights = superlevel_weights / norm(superlevel_weights) #step 2.10
        superlevel_scores = T_scores @ superlevel_weights / (superlevel_weights.T @ superlevel_weights) #step 2.11

        return superlevel_weights, superlevel_scores

    def _y_part(self, superlevel_scores, Y, missing_value_Y):

        if missing_value_Y :
            c_loadings = nansum((Y.T * superlevel_scores).T, axis = 0) 
            c_loadings = c_loadings / nansum(((isnan(Y).T * superlevel_scores) ** 2).T, axis = 0)
        else : c_loadings = Y.T @ superlevel_scores / (superlevel_scores.T @ superlevel_scores) # step 2.12
        
        y_scores = Y @ c_loadings / (c_loadings.T @ c_loadings) # step 2.13

        return c_loadings, y_scores

    def _p_loadings(self, X, scores, missing_value):

        if missing_value:
            p_loadings = nansum(X * scores, axis = 0) 
            p_loadings = array( p_loadings / nansum(scores ** 2), ndmin = 2) #Step 3.1
            
        else: p_loadings = array(X.T @ scores / (scores.T @ scores), ndmin = 2).T #Step 3.1

        return p_loadings

    def _cross_validate(self, X_orig, block_divs, Y_orig, LVs, random_state):

        
        cv_folds = KFold(n_splits = self.cv_splits_number, shuffle = True, random_state = random_state)
        q2 = []

        for train_index, test_index in cv_folds.split(X_orig):
            cv_model = SMB_PLS(tol = self.tol)
            cv_model.fit(X_orig[train_index], block_divs, Y_orig[train_index], latent_variables = LVs, int_call = True)
            q2.append(cv_model.score(X_orig[test_index], Y_orig[test_index]))

        q2 = mean(q2)

        return q2

    def _VIPs_calc( self, X, Y ): # code for calculation of VIPs

        """
        Calculates the VIP scores for all the variables for the prediction
        """
        
        SSY = sum((Y - nanmean(Y, axis = 0)) ** 2)
        for i in range(1, self.x_weights_star.shape[1] + 1):
            pred = self.predict(X, latent_variables = i)
            res = Y - pred
            SSY.loc[i] = sum(((res - res.mean(axis = 0)) ** 2))
            
        SSYdiff = SSY.iloc[:-1] - SSY.iloc[1:]
        VIPs = (((self.x_weights ** 2) @ SSYdiff.values) * self.weights.shape[1] / (SSY[0] - SSY[-1]) ** 1 / 2)
       
        return VIPs


import pandas as pd
from numpy import array, max, sqrt
import pytest
import sys
import os
sys.path.append(".")
from trendfitter.models import SMB_PLS

TESTDATA1_FILENAME = os.path.join(os.path.dirname(__file__), 'smb_pls_dataset.csv')

class TestSMBPLS(object):
    
    def test_fit_missingdataset(self):

        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';')
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        assert test_model.latent_variables == [1, 2], 'Number of principal components is not as expected'

        block_p_loadings = array([[ 0.72187055,  0.53149537,  0.38586444,  0.18720989, -0.14682118, -0.01077914,
                                   -0.19177375,  0.73084692,  0.57978842, -0.30484587],
                                  [ 0.,          0.,          0.,          0.,          0.,          0.,
                                    0.46434217,  0.8654855,   0.21667575, -0.01441398],
                                  [ 0.,          0.,          0.,          0.,          0.,          0.,
                                   -0.31419272,  0.27049653, -0.89933897,  0.14857099]])

        assert test_model.block_p_loadings == pytest.approx(block_p_loadings), 'block p loadings are not as expected'
        
        superlevel_p_loadings = array([[ 0.71366597,  0.52545454,  0.38147881,  0.18508211, -0.14515245, -0.01065663,
                                        -0.02883137,  0.10987593,  0.08716571, -0.0458307 ],
                                       [ 0.,          0.,          0.,          0.,          0.,          0.,
                                         0.46434217,  0.8654855,   0.21667575, -0.01441398],
                                       [ 0.,          0.,          0.,          0.,          0.,          0.,
                                        -0.31419272,  0.27049653, -0.89933897,  0.14857099]])  

        assert test_model.superlevel_p_loadings == pytest.approx(superlevel_p_loadings), 'superlevel p loadings are not as expected'
        
        x_weights = array([[ 0.29286555,  0.2441086,   0.11274504,  0.05083408, -0.01934472, -0.01542236,
                             0.36462124,  0.82820732,  0.13915657, -0.02219426],
                           [ 0.,          0.,          0.,          0.,          0.,          0.,
                             0.43009356,  0.89499372,  0.11828009, -0.0039479 ],
                           [ 0.,          0.,          0.,          0.,          0.,          0.,
                            -0.31481819,  0.27124384, -0.9044671,   0.09620574]])

        assert test_model.x_weights == pytest.approx(x_weights), 'x_weights are not as expected'

        block_weights = array([[ 0.7292824,   0.60786974,  0.28075331,  0.12658505, -0.04817148, -0.03840416,
                                -0.19177375,  0.73084692,  0.57978842, -0.30484587],
                               [ 0.,          0.,          0.,          0.,          0.,          0.,
                                 0.43009356,  0.89499372,  0.11828009, -0.0039479 ],
                               [ 0.,          0.,          0.,          0.,          0.,          0.,
                                -0.31481819,  0.27124384, -0.9044671,   0.09620574]])

        assert test_model.block_weights == pytest.approx(block_weights), 'block_weights are not as expected'

        x_weights_star = array([[ 7.15396313e-01,  5.26728552e-01,  3.82403741e-01,  1.85530860e-01,
                                 -1.45504383e-01, -1.06824664e-02, -8.49794608e-02,  3.63885135e-02,
                                  3.41552710e-02, -3.92117499e-02],
                                [-6.88328818e-02, -5.06799427e-02, -3.67935242e-02, -1.78511176e-02,
                                  1.39999128e-02,  1.02782881e-03,  4.38418356e-01,  8.91300738e-01,
                                  1.15817980e-01,  5.47696385e-03],
                                [ 2.56501258e-02,  1.88855511e-02,  1.37108675e-02,  6.65210291e-03,
                                 -5.21697646e-03, -3.83013724e-04, -2.69695768e-01,  3.68129549e-01,
                                 -8.82709237e-01,  1.46939644e-01]])

        assert test_model.x_weights_star == pytest.approx(x_weights_star), 'x_weights_star are not as expected'

        superlevel_weights = array([[0.98863427, 0.15034055],
                                    [0.,         1.        ],
                                    [0.,         1.        ]])

        assert test_model.superlevel_weights == pytest.approx(superlevel_weights), 'superlevel weights is not as expected'

        c_loadings = array([[1.11716365],
                            [1.65955812],
                            [0.19872565]])

        assert test_model.c_loadings == pytest.approx(c_loadings), 'c_loadings is not as expected'

        expected_q2 = array([0.270841981393722, 0.8939788875287843, 0.9537011992581154])

        assert test_model.q2y == pytest.approx(expected_q2), 'Q2 results are not the expected'
    
    def test_fit_fulldataset(self):

        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        assert test_model.latent_variables == [2, 2], 'Number of principal components is not as expected'

        block_p_loadings = array([[ 0.72434131,  0.54728295,  0.39237194,  0.16893039, -0.16190754, -0.02024584,
                                   -0.04310647,  0.76757319,  0.59608966, -0.23162544],
                                  [-0.05372319,  0.42436286, -0.5471112,  -0.36867179,  0.62294278, -0.09768928,
                                   -0.55470883, -0.27699636,  0.20925143, -0.75616465],
                                  [ 0.,          0.,          0.,          0.,          0.,          0.,
                                    0.46655164,  0.86575441,  0.21084529, -0.01300101],
                                  [ 0.,          0.,          0.,          0.,          0.,          0.,
                                   -0.3170945,   0.26587084, -0.89498554,  0.17547119]])

        assert test_model.block_p_loadings == pytest.approx(block_p_loadings), 'block p loadings are not as expected'
        
        superlevel_p_loadings = array([[ 0.71675602,  0.54155181,  0.38826303,  0.16716135, -0.16021205, -0.02003382,
                                        -0.00622203,  0.11079223,  0.08604014, -0.03343303],
                                      [ -0.05359194,  0.42332612, -0.54577458, -0.3677711,   0.6214209,  -0.09745062,
                                        -0.03875089, -0.01935043,  0.0146179,  -0.05282421],
                                      [  0.,          0.,          0.,          0.,          0.,          0.,
                                         0.46655164,  0.86575441,  0.21084529, -0.01300101],
                                      [  0.,          0.,          0.,          0.,          0.,          0.,
                                        -0.3170945,   0.26587084, -0.89498554,  0.17547119]])

        assert test_model.superlevel_p_loadings == pytest.approx(superlevel_p_loadings), 'superlevel p loadings are not as expected'
        
        x_weights = array([[ 2.89028752e-01,  2.43097117e-01,  1.13597963e-01,  4.38808090e-02,
                            -1.78012004e-02, -1.82581150e-02,  3.73370763e-01,  8.27117856e-01,
                             1.34396517e-01, -1.71408897e-02],
                           [ 1.14588159e-03,  2.90540134e-02, -4.83776317e-02, -2.65612009e-02,
                             5.31168645e-02, -1.16404001e-02,  4.28628544e-01,  8.92442460e-01,
                             1.13832881e-01, -4.22621652e-03],
                           [ 0.00000000e+00,  0.00000000e+00,  0.00000000e+00,  0.00000000e+00,
                             0.00000000e+00,  0.00000000e+00,  4.32168914e-01,  8.94722052e-01,
                             1.12705172e-01,  1.56026869e-04],
                           [ 0.00000000e+00,  0.00000000e+00,  0.00000000e+00,  0.00000000e+00,
                             0.00000000e+00,  0.00000000e+00, -3.16172025e-01,  2.66376764e-01,
                            -9.02463659e-01,  1.20987665e-01]])

        assert test_model.x_weights == pytest.approx(x_weights), 'x_weights are not as expected'

        block_weights = array([[ 7.26868350e-01,  6.11356478e-01,  2.85683564e-01,  1.10354320e-01,
                                -4.47676195e-02, -4.59166979e-02, -4.31064677e-02,  7.67573194e-01,
                                 5.96089659e-01, -2.31625438e-01],
                               [ 1.38468251e-02,  3.51088493e-01, -5.84594961e-01, -3.20965364e-01,
                                 6.41863817e-01, -1.40662512e-01, -5.54708825e-01, -2.76996362e-01,
                                 2.09251432e-01, -7.56164646e-01],
                               [ 0.00000000e+00,  0.00000000e+00,  0.00000000e+00,  0.00000000e+00,
                                 0.00000000e+00,  0.00000000e+00,  4.32168914e-01,  8.94722052e-01,
                                 1.12705172e-01,  1.56026869e-04],
                               [ 0.00000000e+00,  0.00000000e+00,  0.00000000e+00,  0.00000000e+00,
                                 0.00000000e+00,  0.00000000e+00, -3.16172025e-01,  2.66376764e-01,
                                -9.02463659e-01,  1.20987665e-01]])

        assert test_model.block_weights == pytest.approx(block_weights), 'block_weights are not as expected'

        x_weights_star = array([[ 7.15972705e-01,  6.22556924e-01,  2.96930764e-01,  1.04477543e-01,
                                 -5.28212129e-02, -3.74194041e-02, -7.42995455e-02,  3.24027112e-02,
                                  2.92518072e-02, -3.41649738e-02],
                                [ 7.16271143e-02,  5.25517020e-01, -4.86389517e-01, -3.44374029e-01,
                                  6.03391924e-01, -1.02567751e-01, -4.19187192e-02,  1.93774854e-02,
                                  9.78692444e-03, -5.51939484e-02],
                                [-7.35578111e-02, -4.88075861e-02, -4.73880913e-02, -2.23405274e-02,
                                  2.53371106e-02,  6.11775123e-04,  4.38662547e-01,  8.91746733e-01,
                                  1.11027040e-01,  7.83957779e-03],
                                [ 2.98308771e-02,  3.46056564e-02,  2.71565682e-03, -2.28560350e-03,
                                  9.18728002e-03, -3.40803144e-03, -2.73024921e-01,  3.63706701e-01,
                                 -8.78599894e-01,  1.73202743e-01]])

        assert test_model.x_weights_star == pytest.approx(x_weights_star), 'x_weights_star are not as expected'

        superlevel_weights = array([[0.98952802, 0.14434094],
                                    [0.99755694, 0.06985808],
                                    [0.,         1.        ],
                                    [0.,         1.        ]])

        assert test_model.superlevel_weights == pytest.approx(superlevel_weights), 'superlevel weights is not as expected'

        c_loadings = array([[1.13247757],
                            [0.22444752],
                            [1.66790302],
                            [0.19780693]])

        assert test_model.c_loadings == pytest.approx(c_loadings), 'c_loadings is not as expected'

        expected_q2 = array([0.28885562322442215, 0.32110795878481263, 0.9385592222790314, 0.9856995899828982])

        assert test_model.q2y == pytest.approx(expected_q2), 'Q2 results are not the expected'
    
    def test_transform_missingdataset(self):

        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';')
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        scores = array([[ 0.31413744,  0.39130973, -0.11099514],
                        [-0.2680073,  -0.66810093,  0.4425192 ],
                        [-0.28107877, -1.03087832, -0.15124386],
                        [-0.28268446, -0.35992011, -0.31941046],
                        [-0.3982549,   0.19327937, -0.57926287]])

        assert test_model.transform(X_test_data.iloc[:5]) == pytest.approx(scores), 'The transform function malfunctions for datasets with missing values'

    def test_transform_fulldataset(self):
        
        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        scores = array([[ 0.28812785, -0.02374087,  0.38154197, -0.1253679 ],
                        [-0.23870014,  0.0836608,  -0.65833823,  0.47291029],
                        [-0.22974921,  0.32751755, -1.02398531, -0.1502275 ],
                        [-0.31065476, -0.21057959, -0.36306849, -0.29460576],
                        [-0.42839992, -0.31055099,  0.18820115, -0.58135756]])

        assert test_model.transform(X_test_data.iloc[:5]) == pytest.approx(scores), 'The transform function for perfect datasets malfunctions'

    def test_predict(self):

        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        predictions = array([[ 0.93254611],
                             [-1.25604448],
                             [-1.92429957],
                             [-1.06291171],
                             [-0.35595099]])

        assert test_model.predict(X_test_data.iloc[:5]) == pytest.approx(predictions), 'The predict function for perfect datasets malfunctions'
  
    def test_score(self):
        
        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        assert test_model.score(X_test_data.iloc[:5], Y_test_data.iloc[:5]) == pytest.approx(0.9934766558278576), 'The score function for perfect datasets malfunctions'
 
    def test_Hotellings_T2(self):

        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        T2s = array([[1.81687576],
                     [5.34981048],
                     [9.13485717],
                     [2.66708715],
                     [4.84775396]])

        assert test_model.Hotellings_T2(X_test_data.iloc[:5]) == pytest.approx(T2s), 'The Hotellings_T2 function for perfect datasets malfunctions'
"""
    def test_Hotellings_T2_blocks(self): #To Do

        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)  

        #assert test_model.transform(X_test_data.iloc[:5]) == pytest.approx(scores), 'The transform function for perfect datasets malfunctions'

    def test_SPEs_X(self): # To Do
        
        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        #assert test_model.transform(X_test_data.iloc[:5]) == pytest.approx(scores), 'The transform function for perfect datasets malfunctions'

    def test_SPEs_Y(self): # To Do

        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        #assert test_model.transform(X_test_data.iloc[:5]) == pytest.approx(scores), 'The transform function for perfect datasets malfunctions'

    def test_contributions_scores(self): # To Do

        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        #assert test_model.transform(X_test_data.iloc[:5]) == pytest.approx(scores), 'The transform function for perfect datasets malfunctions'
    
    def test_contributions_scores_block(self): # To Do
        
        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        #assert test_model.transform(X_test_data.iloc[:5]) == pytest.approx(scores), 'The transform function for perfect datasets malfunctions'

    def test_contributions_SPE_X(self): # To Do
        
        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        #assert test_model.transform(X_test_data.iloc[:5]) == pytest.approx(scores), 'The transform function for perfect datasets malfunctions'

    def test_contributions_SPE_X_block(self): # To Do
        
        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0, delimiter = ';').dropna()
        test_data = (test_data - test_data.mean()) / test_data.std()
        X_test_data = test_data.drop(columns = 'y')
        X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] = X_test_data[['Var1', 'Var2', 'Var3', 'rand1', 'rand2', 'rand3']] / sqrt(6)
        X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] = X_test_data[['Var4', 'Var5', 'rand4', 'rand5']] / sqrt(4)
        Y_test_data = test_data['y']
        test_model = SMB_PLS(tol = 1e-8, cv_splits_number = 7)
        test_model.fit(X_test_data, [6,10], Y_test_data, random_state = 2)

        #assert test_model.transform(X_test_data.iloc[:5]) == pytest.approx(scores), 'The transform function for perfect datasets malfunctions'
    """
    

    




from numpy import min, sum, mean, std, nanvar, insert, array, where, isnan, nan_to_num, \
                    nansum, nanmean, identity, zeros, unique, var, append, nansum, any, \
                    concatenate
from numpy.core.numeric import ones, outer
from numpy.linalg import norm, pinv
from pandas import DataFrame, Series
from sklearn.model_selection import KFold
from scipy.stats import f, chi2
from trendfitter.auxiliary.tf_aux import scores_with_missing_values

class DiPLS:

    """
    A sklearn-like class for the Projection to Latent Structures.

        Parameters
        ----------
        cv_splits_number : int, optional
            number of splits used for cross validation in case latent_variables is None

        Attributes
        ----------
        latent_variables : int
            number of latent variables one wants to extract.
        
        Methods
        -------
        fit(X, Y, latent_variables = None, deflation = 'both', random_state = None, int_call = False)
            Applies the NIPALS like method to find the best parameters that adjust the model 
                to the data 
    
    """

    def __init__(self, cv_splits_number = 7, tol = 1e-8, loop_limit = 1000, missing_values_method = 'TSM'):
        
        self.p_loadings = None
        self.weights = None
        self.weights_star = None
        self.c_loadings = None
        self.latent_variables = None # number of principal components to be extracted
        self.cv_splits_number = cv_splits_number # number of splits for latent variable cross-validation
        self.tol = tol # criteria for convergence
        self.loop_limit = loop_limit # maximum number of loops before convergence is decided to be not attainable
        self.q2y = [] # list of cross validation scores
        self.deflation = None
        self.coefficients = None
        self.training_scores = None
        self.omega = None
        self.missing_values_method = missing_values_method
        self.feature_importances_ = None #for scikit learn use with feature selection methods
        self._x_chi2_params = []
        self._y_chi2_params = []

    def fit(self, X, Y, latent_variables = None, s = 0, deflation = 'both', random_state = None, int_call = False):

        """
        Adjusts the model parameters to best fit the Y using the algorithm defined in 
            Dong et Qin's [1]

        Parameters
        ----------
        X : array_like
            Matrix with all the data to be used as predictors in one only object
        Y : array_like
            Matrix with all the data to be predicted in one only object
        latent_variables : int, optional
            number of latent variables deemed relevant. If left unspecified
                a cross validation routine will define the number during fitting.
        s : int

        deflation : str, optional
            string defining method of deflation, only Y or both X and Y 
        random_state : int, optional
            value used as a seed in the random number generator for cross validation

        References 
        ----------
        [1] Y. Dong and S. J. Qin, “Regression on dynamic PLS structures for supervised 
        learning of dynamic data,” Journal of Process Control, vol. 68, pp. 64–72, 
        Aug. 2018, doi: 10.1016/j.jprocont.2018.04.006.

        """

        """----------------------- Dealing with data possibly coming in the form of a pandas dataframe ------------------------"""
        
        self.latent_variables = latent_variables
        self.s = s
        self.deflation = deflation
        
        if isinstance(X, DataFrame): # in case the function receives a dataframe as X data
            X = array(X.to_numpy(), ndmin = 2)
        Orig_X = X
            
        if isinstance(Y, DataFrame): Y = array( Y.to_numpy(), ndmin = 2)
        elif isinstance(Y, Series): Y = array( Y.to_numpy(), ndmin = 2).T # in case the function receives a dataframe as Y data
        else: Y = array(Y, ndmin = 2)
        Orig_Y = Y

        G = identity(X.shape[1])

        for latent_variable in range(1, latent_variables + 1):
            
            """-------------------------------- Model Building Loop ---------------------------------------"""
            y_scores, y_scores_hat, x_scores, weights, q_loadings, outer_beta, inner_beta, outer_alpha, inner_alpha = self._1LV_loop(X, Y, s)
            p_loadings = (X.T @ x_scores / (x_scores.T @ x_scores)).T


            """-------------------------------- Deflation -------------------------------------------------"""
            X -= x_scores @ p_loadings
            Y[s:, :]-= y_scores_hat @ q_loadings.T


            """-------------------------------- class property assignment section -------------------------"""
            if latent_variable < 2:
                self.p_loadings = p_loadings
                self.weights = weights
                self.q_loadings = q_loadings
                self.weights_star = weights
                self.outer_beta = outer_beta
                self.inner_beta = inner_beta
                self.outer_alpha = outer_alpha
                self.inner_alpha = inner_alpha
                G = G - weights.T @ p_loadings 
                self.training_x_scores = array(x_scores, ndmin = 2)

            else:
                weights_star = array(G @ weights.T , ndmin = 2).T
                G = G - weights_star.T @ p_loadings 
                self.weights_star = insert(self.weights_star, self.weights_star.shape[0], weights_star, axis = 0)          
                self.p_loadings = insert(self.p_loadings, self.p_loadings.shape[0], p_loadings, axis = 0)
                self.weights = insert(self.weights, self.weights.shape[0], weights, axis = 0)
                self.q_loadings = insert(self.q_loadings, self.q_loadings.shape[0], q_loadings, axis = 0)
                self.outer_beta = insert(self.outer_beta, self.outer_beta.shape[0], outer_beta, axis = 0)
                self.inner_beta = insert(self.inner_beta, self.inner_beta.shape[0], inner_beta, axis = 0)
                self.outer_alpha = insert(self.outer_alpha, self.outer_alpha.shape[0], outer_alpha, axis = 0)
                self.inner_alpha = insert(self.inner_alpha, self.inner_alpha.shape[0], inner_alpha, axis = 0)
                self.training_x_scores = insert(self.training_x_scores, self.training_x_scores.shape[1], array( x_scores, ndmin = 2).T, axis = 1)
                
            self.omega = self.training_x_scores.T @ self.training_x_scores 

        pass

    def transform_to_y_scores(self, X, latent_variables = None):

        if isinstance(X, DataFrame): X = X.to_numpy()      
        if latent_variables is None: latent_variables = self.latent_variables

        t_scores = X @ self.weights_star[:latent_variables, :].T

        return t_scores

    def transform_di(self, X, old_Ys = None, latent_variables = None, s = None):

        if isinstance(X, DataFrame): X = X.to_numpy()      
        if latent_variables is None: latent_variables = self.latent_variables
        if s is None: s = self.s
        
        if old_Ys is None:
            t_scores = self.transform_to_y_scores(X, latent_variables = latent_variables)
            y_len = t_scores.shape[0]
            y_scores_hat = sum([t_scores[(s - i):y_len - i, :] * beta for i, beta in enumerate(self.inner_beta.T)], axis = 0)
        else:
            if isinstance(old_Ys, DataFrame): old_Ys = old_Ys.to_numpy()
            t_scores = self.transform_to_y_scores(X, latent_variables = latent_variables)
            y_len = t_scores.shape[0]
            y_scores_hat = sum([t_scores[(s - i):y_len - i, :] * beta for i, beta in enumerate(self.inner_beta.T)], axis = 0)
            y_scores = old_Ys @ self.q_loadings.T
            y_scores_hat += sum([y_scores[(s - (i + 1)):y_len - (i + 1), :] * beta for i, beta in enumerate(self.inner_alpha.T)], axis = 0)


        return y_scores_hat

    def predict(self, X, old_Ys = None, latent_variables = None, s = None):
    
        y_scores_hat = self.transform_di(X, old_Ys = old_Ys, latent_variables = latent_variables, s = s)
        
        preds = y_scores_hat @ self.q_loadings

        return preds
        
    def score(self, X, Y, latent_variables = None, s = None):

        if isinstance(Y, DataFrame) or isinstance(Y, Series): Y = array(Y.to_numpy(), ndmin = 2).T  
        if latent_variables is None: latent_variables = self.latent_variables
        if s is None: s = self.s 

        Y_hat = self.predict(X, Y, latent_variables = latent_variables, s = s)
        F = Y[s:, :] - Y_hat
        score = 1 - nanvar(F) / nanvar(Y)
        
        return score

    def _1LV_loop(self, X, Y, s):

        y_scores, x_scores, weights, q_loadings, outer_beta, outer_alpha = self._outer_loop(X, Y, s)

        y_scores_hat, inner_beta, inner_alpha = self._inner_loop(x_scores, y_scores, s)

        return y_scores, y_scores_hat, x_scores, weights, q_loadings, outer_beta.T, inner_beta.T, outer_alpha.T, inner_alpha.T
    
    def _outer_loop(self, X, Y, s):
        
        outer_alpha = array([0 for i in range(s)], ndmin = 2).T
        if len(outer_alpha>0): outer_alpha[0] = 1
        outer_beta = array([1] + [0 for i in range(s)], ndmin = 2).T
        y_scores = array(Y[:, 0], ndmin = 2).T
        q_loadings = ones((1, Y.shape[1]))
        conv = 1
        loops = 0
        X_len = X.shape[0]

        while (conv > self.tol and loops < self.loop_limit):
            
            
            x_beta = sum([X[(s - i):X_len - i, :] * beta for i, beta in enumerate(outer_beta)], axis = 0).T
            weights = array(x_beta @ y_scores[s:], ndmin = 2).T
            weights /= norm(weights)
            x_scores = X @ weights.T
            t_beta = sum([x_scores[s - i:X_len - i] * beta for i, beta in enumerate(outer_beta)], axis = 0)
            y_alpha = sum([Y[(s - (i + 1)): X_len - (i + 1), :] * alpha for i, alpha in enumerate(outer_alpha)], axis = 0)
            u_alpha = y_alpha @ q_loadings.T

            q_loadings_calc = Y[s:, :].T @ t_beta + y_alpha.T @ y_scores[s:, :] + Y[s:, :].T @ u_alpha
            #q_loadings = Y[s:, :].T @ t_beta versão funcional
            q_loadings_calc /= norm(q_loadings_calc)


            y_scores_calc = Y @ q_loadings.T
            
            T_s = concatenate([x_scores[(s - i):X_len - i, :] for i in range(s + 1)], axis = 1)
            outer_beta_calc = T_s.T @ y_scores_calc[s:, :]
            outer_beta_calc /= norm(outer_beta_calc)

            U_s = concatenate([y_scores[(s - i):X_len - i, :] for i in range(s)], axis = 1)
            outer_alpha_calc = U_s.T @ y_scores_calc[s:, :]
            outer_alpha_calc /= norm(outer_alpha_calc)
             

            loops += 1
            conv = norm(y_scores - y_scores_calc) + norm(outer_beta - outer_beta_calc) + \
                    norm(outer_alpha - outer_alpha_calc) + norm(q_loadings - q_loadings_calc)
            y_scores = y_scores_calc
            outer_beta = outer_beta_calc
            outer_alpha = outer_alpha_calc
            q_loadings = q_loadings_calc

        return y_scores_calc, x_scores, weights, q_loadings, outer_beta, outer_alpha

    def _inner_loop(self, x_scores, y_scores, s):

        X_len = x_scores.shape[0]
        big_Ts = concatenate([x_scores[(s - i):X_len - i, :] for i in range(s+1)] + \
                             [y_scores[(s - (i + 1)):- (i + 1)] for i in range(s)], axis = 1)
        inner_beta_alpha = (pinv(big_Ts.T @ big_Ts) @ big_Ts.T) @ y_scores[s:]
        y_scores_hat = big_Ts @ inner_beta_alpha

        return y_scores_hat, inner_beta_alpha[:s+1, :], inner_beta_alpha[s+1:2 * s + 1 , :]





from numpy import min, sum, mean, std, var, insert, array, multiply, where, zeros, \
                  append, isnan, nan_to_num, nansum, nanvar, nanmean, unique, ix_, \
                  nonzero, nan, diagonal, sqrt
from numpy.linalg import norm
from scipy.stats import f, chi2
from pandas import DataFrame
from sklearn.model_selection import KFold
from trendfitter.auxiliary.tf_aux import scores_with_missing_values
from . import PCA

class MB_PCA:
    """
    A sklearn-like class for the Multi-Block Principal Component Analysis.

        Parameters
        ----------
        cv_splits_number : int, optional
            number of splits used for cross validation in case latent_variables is None
        tol : float, optional
            value used to decide if model has converged
        loop_limit : int, optional
            maximum number of loops for the extraction of one latent variable
        missing_values_method : str, optional
            string to define the method that the model will deal with missing values

        Attributes
        ----------
        principal_components : int, optional
            Number of principal components extracted or to extract. If not given, a cross-validation
                internal routine will decide this value.
        loadings : array_like
            Loading parameters that define the PCA model.
        q2 : [float]
            list of r² coefficients extracted during cross validation        
        training_scores : array_like
            Scores extracted during training of the model.
        superlever_training_scores : array_like
            Scores extracted during training of the model for the superlevel matrix.
        omega : array_like
            If missing_values_method requires a scores covariance matrix ('TSR', 'CMR', 'PMP'), 
                it will be stored here. 
        omega_b : [array_like]
            If missing_values_method requires a scores covariance matrix ('TSR', 'CMR', 'PMP'), 
                the matrixes for each block will be stored here. 
        block_divs : [int]
            list with the index of every block final position. ex. [2,4] means two blocks with 
                columns 0 and 1 on the first block and columns 2 and 3 on the second block. 
                Assigned when fit method runs.
        block_loadings : [array_like]
            list of p loadings arrays of every block with all the extracted latent variables 
        superlevel_loadings : array_like
            array of all latent variables extracted p loadings for the super level
        feature_importances_ : [float]
            list of values that represent how important is each variable in the same order 
                of the X columns on the first matrix

        Methods
        -------
        fit(X, blocks_sep)
            Applies the NIPALS like method to find the best parameters that adjust the model 
                to the data
        transform(X)
            Transforms the original data to the latent variable space in the super level 
        transform_b(X, block)
            Transforms the original data to the latent variable space in the block level for
                all blocks
        predict(X)
            Predicts Y values 
        score(X)
            Returns a r² representing how much variability from X is captured in the model.
        Hotellings_T2(X)
            Calculates Hotellings_T2 values for the X data in the super level
        T2_limit(alpha)
           Returns the Hotelling's T² limit estimated with alpha confidence level
        SPEs(X)
            Calculates squared prediction errors on the X side for the super level
        SPE_limit(alpha)
            Returns the squared prediction error limit with alpha confidence level
        contributions_scores_ind(X, principal_components = None)
            Returns an array with the contributions to the scores of each X row.
        contributions_spe(X)
            calculates the contributions of each variable to the SPE on the X side for
                the super level

    """
    def __init__(self, cv_splits_number = 7, tol = 1e-16, loop_limit = 1000, missing_values_method = 'TSM'):
        
        # Parameters
        
        self.cv_splits_number = cv_splits_number # number of splits for latent variable cross-validation
        self.tol = tol # criteria for convergence
        self.loop_limit = loop_limit # maximum number of loops before convergence is decided to be not attainable        
        self.missing_values_method = missing_values_method
        
        # Attributes

        self.principal_components = None # number of principal components to be extracted
        self.loadings = None
        self.q2 = [] # list of cross validation scores
        self.block_divs = None
        self.block_loadings = None
        self.superlevel_loadings = None
        self.omega = None # score covariance matrix for missing value estimation
        self.omega_b = []
        self.training_scores = None
        self.superlevel_training_scores = None
        self.feature_importances_ = None
        self._chi2_params = None
        self._pca = None
        
    def fit(self, X, block_divs, principal_components = None, deflation = 'both', int_call = False): #require test

        """
        Adjusts the model parameters to best fit the Y using the algorithm defined in 
            Westerhuis et al's [1]

        Parameters
        ----------
        X : array_like
            Matrix with all the data to be used as predictors in one only object
        block_divs : [int]
            list with the index of every block final position. ex. [2,4] means two blocks with 
                columns 0 and 1 on the first block and columns 2 and 3 on the second block.
        Y : array_like
            Matrix with all the data to be predicted in one only object
        latent_variables : int, optional
            Number of latent variables deemed relevant from each block. If left unspecified
                a cross validation routine will define the number during fitting
        deflation : str, optional
            string defining method of deflation, only Y or both X and Y 
        int_call : Boolean, optional
            Flag to define if it is an internal call on the cross validation routine and decide
                if it is necessary to calculate the VIP values

        References 
        ----------
        [1] J. A. Westerhuis, T. Kourti, and J. F. MacGregor, “Analysis of multiblock and 
        hierarchical PCA and PLS models,” Journal of Chemometrics, vol. 12, no. 5, pp. 301–321, 
        1998, doi: 10.1002/(SICI)1099-128X(199809/10)12:5<301::AID-CEM515>3.0.CO;2-S.

        """
        if isinstance(X, DataFrame):# If X data is a pandas Dataframe
            X = array(X.to_numpy(), ndmin = 2)

        self.block_divs = block_divs
        block_coord_pairs = (*zip([0] + block_divs[:-1], block_divs),)
        
        #Using a full PLS as basis to calculate the MB-PLS
        int_PCA = PCA(tol = self.tol, loop_limit = self.loop_limit, missing_values_method = self.missing_values_method)
        int_PCA.fit(X, principal_components = principal_components,  cv_splits_number = self.cv_splits_number)

        

        block_coord_pairs = (*zip([0] + block_divs[:-1], block_divs),)

        superlevel_T = zeros((X.shape[0], len(block_coord_pairs) * int_PCA.principal_components))
        block_loadings = zeros((int_PCA.principal_components, X.shape[1]))

        for block, (start, end) in enumerate(block_coord_pairs):
            test_missing_data = isnan(sum(X[:, start:end]))
            if test_missing_data:
                b_loadings = zeros((int_PCA.principal_components, end - start))
                for i in range(int_PCA.principal_components):
                    b_loadings[i, :] = nansum(X[:, start:end] * array(int_PCA.training_scores[:, i], ndmin = 2).T, axis = 0) / nansum(((~isnan(X[:, start:end]).T * int_PCA.training_scores[:, i]) ** 2), axis = 1)
            else:
                b_loadings = array(X[:, start:end].T @ int_PCA.training_scores / diagonal(int_PCA.training_scores.T @ int_PCA.training_scores), ndmin = 2).T
            
            block_loadings[:int_PCA.principal_components, start:end] = b_loadings
            block_scores = zeros((X.shape[0], int_PCA.principal_components))
            if test_missing_data:
                for i in range(int_PCA.principal_components):
                    block_scores[:, i] = nansum(X[:, start:end] * b_loadings[i, :], axis = 1) / nansum(((~isnan(X[:, start:end]) * b_loadings[i, :]) ** 2), axis = 1)
            else:
                block_scores = (X[:, start:end] @ b_loadings.T)

            superlevel_T[:, [block + num * len(block_coord_pairs) for num, _ in enumerate(block_coord_pairs)]] = block_scores
            self.omega_b.append(block_scores.T @ block_scores)

 
        superlevel_loadings = array(superlevel_T.T @ int_PCA.training_scores / diagonal(int_PCA.training_scores.T @ int_PCA.training_scores), ndmin = 2).T

        #----------------Attribute Assignment---------------

        self.principal_components = int_PCA.principal_components
        self.block_divs = block_divs 
        self.block_loadings = block_loadings #
        self.superlevel_loadings = superlevel_loadings #
        self.superlevel_training_scores = superlevel_T #
        self.training_scores = int_PCA.training_scores
        self.loadings = int_PCA.loadings 
        self.q2 = int_PCA.q2


        self.feature_importances_ = int_PCA.feature_importances_
        self.omega = int_PCA.omega
        self._chi2_params = int_PCA._chi2_params
        self._pca = int_PCA

    def predict(self, X, principal_components = None):

        """
        Transforms the X sample to the principal component space and back to evaluate what is
        the model "prediction" of the original sample values.

        Parameters
        ----------
        X : array_like
            Samples to reproduce
        principal_components : array_like, Optional
            Number of desired principal components to be used
        
        Returns
        -------
        preds : array_like 
            returns "predicted" X values.        
        """

        return self._pca.predict(X, principal_components = principal_components)
    
    def transform(self, X, principal_components = None): 
        
        """

        Transforms the X sample to the principal component space .

        Parameters
        ----------
        X : array_like
            Samples to transform
        principal_components : array_like, Optional
            Number of desired principal components to be used
        
        Returns
        -------
        result : array_like 
            returns X samples' scores.  

        """

        return self._pca.transform(X, principal_components)
    
    def score(self, X, principal_components = None): #return r²
        
        """
        
        Returns the coefficient of determination R^2 of the model.

        R² is defined as 1 - Variance(Error) / Variance(X) with Error = X - predictions(X)

        Parameters
        ----------
        X : array_like
            Samples used for score calculation
        principal_components : array_like, Optional
            Number of desired principal components to be used
        
        Returns
        -------
        result : array_like 
            returns calculated r².        
        
        """
        return self._pca.score(X, principal_components = principal_components)

    def Hotellings_T2(self, X, principal_components = None):
        
        """
        Calculates the Hotelling's T² for the X samples.

        Parameters
        ----------
        X : array_like
            Samples Matrix
        principal_components : int, optional
            number of latent variables to be used. 

        Returns
        -------
        T2s : array_like 
            returns all calculated T²s for the X samples
        """

        return self._pca.Hotellings_T2(X, principal_components = principal_components)
    
    def T2_limit(self, alpha, principal_components = None):
        
        """
        Calculates the Hotelling's T² limit based on the training dataset.

        Parameters
        ----------
        alpha : array_like
            value ranging from 0-1 to represent the % confidence limit
        principal_components : int, optional
            number of latent variables to be used. 

        Returns
        -------
        t2_limit : array_like 
            returns the limit T² for the alpha based on the training dataset
        """

        return self._pca.T2_limit(alpha, principal_components = principal_components)
    
    def contributions_scores_ind(self, X, principal_components = None) : #contribution of each individual point in X1
        
        """
        calculates the sample individual contributions to the scores.

        Parameters
        ----------
        X : array_like
            Sample matrix 
        principal_components : int, optional
            number of latent variables to be used. 

        Returns
        -------
        contributions : array_like 
            matrix of scores contributions for every X sample
        """

        return self._pca.contributions_scores_ind(X, principal_components = principal_components)
       
    def contributions_spe(self, X, principal_components = None) :
        
        """
        calculates the individual sample individual contributions to the squared prediction error 
        of the X variables matrix reconstruction.

        Parameters
        ----------
        X : array_like
            Sample matrix 
        principal_components : int, optional
            number of latent variables to be used. 

        Returns
        -------
        SPE_contributions : array_like 
            matrix of SPE contributions for every X sample
        """
        
        return self._pca.contributions_spe(X, principal_components = principal_components)   

    def SPEs(self, X, principal_components = None) :
        
        """
        Calculates the Squared prediction error for for the X matrix rebuild.

        Parameters
        ----------
        X : array_like
            Samples Matrix
        principal_components : int, optional
            number of latent variables to be used. 

        Returns
        -------
        SPE : array_like 
            returns all calculated SPEs for the X samples
        """
        
        return self._pca.SPEs(X, principal_components = principal_components)

    def SPE_limit(self, alpha, principal_components = None):

        """
        Calculates the SPE limit based on the training dataset.

        Parameters
        ----------
        alpha : array_like
            value ranging from 0-1 to represent the % confidence limit
        principal_components : int, optional
            number of latent variables to be used. 

        Returns
        -------
        SPE_limit : array_like 
            returns the limit SPE for the alpha based on the training dataset
        """
        return self._pca.SPE_limit(alpha, principal_components = principal_components)

    def transform_b(self, X, block_index, principal_components = None): #require test

        """

        Transforms the X sample to the latent space on the chosen block level.

        Parameters
        ----------
        X : array_like 
            Samples to transform to the block scores. dim 1 of the array should be equal to the block's.
        block_index : int
            Desired block starting from zero
        principal_components : array_like, Optional
            Number of desired principal components to be used
        
        Returns
        -------
        result : array_like 
            returns X samples' scores for the chosen block.  

        """

        if isinstance( X, DataFrame ): X = X.to_numpy()            
        if principal_components == None: principal_components = self.principal_components
        
        if block_index == 0: start, end = 0, self.block_divs[block_index]
        else: start, end = self.block_divs[block_index], self.block_divs[block_index + 1]

        if isnan(sum(X)):
            result = zeros((X.shape[0], principal_components))
            X_nan = isnan(X)
            variables_missing_mask = unique(X_nan, axis = 0)
            
            for row_mask in variables_missing_mask:
                
                rows_indexes = where((X_nan == row_mask).all(axis = 1))                
                
                if sum(row_mask) == 0: 

                    result[rows_indexes, :] = X[rows_indexes, :] @ self.loadings[:principal_components, start:end].T 
                
                else:
                    
                    result[rows_indexes, :] = scores_with_missing_values(self.omega_b[block_index], self.block_loadings[:, start:end][:, ~row_mask], X[rows_indexes[0][:, None], ~row_mask], 
                                                                            LVs = principal_components, method = self.missing_values_method)
                 
                    
        else: result = X @ self.block_loadings[:principal_components, start:end].T

        return result

    def predict_b(self, X, block_index, principal_components = None): #require test

        """

        Transforms the X sample to the latent space on the chosen block level.

        Parameters
        ----------
        X : array_like 
            Samples to rebuild using the block scores. dim 1 of the array should be equal to the block's.
        block_index : int
            Desired block starting from zero
        principal_components : array_like, Optional
            Number of desired principal components to be used
        
        Returns
        -------
        result : array_like 
            returns X samples rebuilt using the model's block loadings.  

        """
         
        if principal_components == None: principal_components = self.principal_components
        
        if block_index == 0: start, end = 0, self.block_divs[block_index]
        else: start, end = self.block_divs[block_index], self.block_divs[block_index + 1]

        b_scores = self.transform_b(X, block_index, principal_components = principal_components)

        result = b_scores @ self.block_loadings[:principal_components, start:end]

        return result

    def score_b(self, X, block_index, principal_components = None): #require test

        """
        
        Returns the coefficient of determination R^2 referring to a specific block.
        dim 1 of the array should be equal to the block's.

        R² is defined as 1 - Variance(Error) / Variance(X) with Error = X - predictions(X)

        Parameters
        ----------
        X : array_like
            Samples used for score calculation
        block_index : int
            Desired block starting from zero
        principal_components : array_like, Optional
            Number of desired principal components to be used
        
        Returns
        -------
        result : array_like 
            returns calculated r².        
        
        """
        if isinstance(X, DataFrame) : X = X.to_numpy()
        
        error = X - self.predict_b(X, block_index, principal_components = principal_components)
        result = 1 - nanvar(error) / nanvar(X)
        return result



from numpy import array, isnan, nansum, nan_to_num, multiply, sum, sqrt, append, zeros, place, nan, concatenate, mean, nanvar, std, unique, where, nanmean, identity
from numpy.linalg import norm, pinv
from sklearn.model_selection import KFold
from pandas import DataFrame, Series
from trendfitter.auxiliary.tf_aux import scores_with_missing_values

class MLSMB_PLS:
    """
    A sklearn-like class for the Sequential Multi-Block Projection to Latent Structures.

        Parameters
        ----------
        cv_splits_number : int, optional
            number of splits used for cross validation in case latent_variables is None
        tol : float, optional
            value used to decide if model has converged
        loop_limit : int, optional
            maximum number of loops for the extraction of one latent variable
        missing_values_method : str, optional
            string to define the method that the model will deal with missing values

        Attributes
        ----------
        latent_variables : [int]
            list of number of latent variables deemed relevant from each block. 
        block_divs : [int]
            list with the index of every block final position. ex. [2,4] means two blocks with 
                columns 0 and 1 on the first block and columns 2 and 3 on the second block. 
                Assigned when fit method runs.
        p_loadings_block : [array_like]
            list of p loadings arrays of every block with all the extracted latent variables 
        superlevel_p_loadings : array_like
            array of all latent variables extracted p loadings for the super level
        weights_block : [array_like]
            list of weights arrays of every block with all the extracted latent variables 
        weights_super : array_like
            array of all latent variables extracted weights for the super level
        c_loadings : array_like
            array of c_loadings for all extracted latent variables
        q2 : [float]
            list of r² coefficients extracted during cross validation
        feature_importances : [float]
            list of values that represent how important is each variable in the same order 
                of the X columns on the first matrix

        Methods
        -------
        fit(X, blocks_sep, Y)
            Applies the NIPALS like method to find the best parameters that adjust the model 
                to the data
        transform(X)
            Transforms the original data to the latent variable space in the super level 
        transform_inv(scores)
            Returns the superlevel scores to the original data space
        transform_b(X, block)
            Transforms the original data to the latent variable space in the block level for
                all blocks
        predict(X)
            Predicts Y values 
        score(X, Y)
            calculates the r² value for Y
        Hotellings_T2(X)
            Calculates Hotellings_T2 values for the X data in the super level
        Hotellings_T2_blocks(X)
            Calculates Hotellings_T2 values for in the block level for all blocks
        SPEs_X(X)
            Calculates squared prediction errors on the X side for the super level
        SPEs_X_blocks(X)
            Calculates squared prediction errors for the block level for all blocks
        SPEs_Y(X, Y)
            Calculates squared prediction errors for the predictions
        contributions_scores(X)
            calculates the contributions of each variable to the scores on the super level
        contributions_scores_b(X)
            calculates the contributions of each variable to the scores on the super level
        contributions_SPE(X)
            calculates the contributions of each variable to the SPE on the X side for
                the super level
        contributions_SPE_b(X)
            calculates the contributions of each variable to the SPE on the X side for
                the block level for all blocks

    """
    def __init__(self, cv_splits_number = 7, tol = 1e-16, loop_limit = 1000, missing_values_method = 'TSM'):
        
        # Parameters
        
        self.cv_splits_number = cv_splits_number # number of splits for latent variable cross-validation
        self.tol = tol # criteria for convergence
        self.loop_limit = loop_limit # maximum number of loops before convergence is decided to be not attainable
        self.q2y = [] # list of cross validation scores
        self.missing_values_method = missing_values_method
        
        # Attributes

        self.latent_variables = None # number of principal components to be extracted
        self.third_level_divs = None
        self.second_level_divs = None
        self.p_loadings_3 = None
        self.p_loadings_2 = None
        self.p_loadings_1 = None
        self.weights_3 = None
        self.weights_2 = None
        self.weights_1 = None
        self.x_weights_star = None
        self.x_weights = None
        self.c_loadings = None 
        self.VIPs = None
        self.coefficients = None
        self.omega = None # score covariance matrix for missing value estimation
        self.training_scores_3 = None
        self.training_scores_2 = None
        self.training_scores_1 = None
        self.training_scores_y = None
            
    def fit(self, X, third_level_divs, second_level_divs, Y, latent_variables = None, deflation = 'both', int_call = False, random_state = None):
        """
        Adjusts the model parameters to best fit the Y using the algorithm defined in 
            

        Parameters
        ----------
        X : array_like
            Matrix with all the data to be used as predictors in one only object
        block_divs : [int]
            list with the index of every block final position. ex. [2,4] means two blocks with 
                columns 0 and 1 on the first block and columns 2 and 3 on the second block.
        Y : array_like
            Matrix with all the data to be predicted in one only object
        latent_variables : [int], optional
            list of number of latent variables deemed relevant from each block. If left unspecified
                a cross validation routine will define the number during fitting
        deflation : str, optional
            string defining method of deflation, only Y or both X and Y 
        int_call : Boolean, optional
            Flag to define if it is an internal call on the cross validation routine and decide
                if it is necessary to calculate the VIP values

        References 
        ----------
        

        """

        
        # check if X and Y make sense
        # check if block_divs make sense 
        # check if latent_variables make sense


        self.third_level_divs = third_level_divs
        self.second_level_divs = second_level_divs
        third_level_pairs = (*zip([0] + third_level_divs[:-1], third_level_divs),)
        second_level_pairs = (*zip(second_level_divs[:-1], second_level_divs[1:]),)

        if isinstance(X, DataFrame):# If X data is a pandas Dataframe
            X = array(X.to_numpy(), ndmin = 2)
          
        if isinstance(Y, DataFrame) or isinstance(Y, Series): # If Y data is a pandas Dataframe or Series 
            Y = array(Y.to_numpy(), ndmin = 2).T
            
        Orig_X = X.copy()
        Orig_Y = Y.copy()
        
        """------------------- model Calculation -------------"""

        all_blocks = set(third_level_pairs).union(set(second_level_pairs))
        max_LVs = [min([pair[1] - pair[0] for pair in third_level_pairs])] + [pair[1] - pair[0]  for pair in second_level_pairs]

        for block, LVs in enumerate(max_LVs):
            for LV in range(LVs):
                if LV >= latent_variables[block]: break

                if block == 0: 
                    result = self._MLSMBPLS_1LV(X,third_level_pairs, second_level_pairs, Y, 0)
                else:
                    missing_values_list = [0 for _ in second_level_pairs]
                    result = self._SMBPLS_1LV(X, second_level_pairs[block-1:], Y, missing_values_list, block)

                
                #------------------------------------deflation----------------------------

                if deflation == 'both' :
                    X -= result['t1l'] @ result['p1l']
                    Y -= result['t1l'] @ result['c']
                elif deflation == 'Y':
                    Y -= result['t1l'] @ result['c']
                else : raise Exception(f'Deflation method "{deflation}" non-existing.')
                    
                #--------------------------Property assignment section--------------------
                if LV == 0 and block == 0 :
                    self.superlevel_weights = result['w1l']
                    #self.x_weights = result['w']
                    #self.block_weights = result['wb']
                    #self.block_p_loadings = result['pb']
                    self.superlevel_p_loadings = result['p1l']
                    self.c_loadings = result['c']
                    self.training_superlevel_scores = result['t1l']
                    #self.training_block_scores = result['tb']
                    self.training_y_scores = result['u']
                    #self.x_weights_star2 = result['w']
                    #for bl, (start, end) in enumerate(block_coord_pairs):
                        #G[bl] = G[bl] - result['w'][:, start:end].T @ result['p'][:, start:end]

                
                else:                    
                    self.superlevel_weights = append(self.superlevel_weights, result['w1l'], axis = 0)
                    #self.x_weights = append(self.x_weights, result['w'], axis = 0)
                    #self.block_weights = append(self.block_weights, result['wb'], axis = 0)
                    #self.block_p_loadings = append(self.block_p_loadings, result['pb'], axis = 0)
                    self.superlevel_p_loadings = append(self.superlevel_p_loadings, result['p1l'], axis = 0)
                    self.c_loadings = append(self.c_loadings, result['c'], axis = 0)
                    self.training_superlevel_scores = append(self.training_superlevel_scores, result['t1l'], axis = 1)
                    #self.training_block_scores = append(self.training_block_scores, result['tb'], axis = 1)
                    self.training_y_scores = append(self.training_y_scores, result['u'], axis = 1)

                    #for bl, (start, end) in enumerate(block_coord_pairs[block:]):
                        #if bl == 0: x_weights_star2 = array(G[block + bl] @ result['w'][:, start:end].T, ndmin = 2).T
                        #else: x_weights_star2 = append(x_weights_star2, array(G[block + bl] @ result['w'][:, start:end].T, ndmin = 2 ).T, axis = 1)
                        #if block > 0 and bl == 0: x_weights_star2 = append(zeros((1, start)), x_weights_star2, axis = 1)

                        #G[block + bl] = G[block + bl] - result['w'][:, start:end].T @ result['p'][:, start:end]
                    
                    #self.x_weights_star2 = append(self.x_weights_star2, x_weights_star2, axis = 0)
                    
        #self.block_weights_star = (self.block_weights @ pinv(self.block_p_loadings.T @ self.block_weights))
        #self.x_weights_star = (self.x_weights @ pinv(self.superlevel_p_loadings.T @ self.x_weights))

        return

    def transform(self, X, latent_variables = None): 
        
        """
        Transforms the X matrix to the model-fitted space returning scores
        of the super level.

        Parameters
        ----------
        X : array_like
            Matrix with all the data to be used as predictors in one only object
        latent_variables : [int], optional
            list with number of latent variables to be used. 

        Returns
        -------
        result : array_like of shape (X.shape[0], sum(latent_variables))
            Scores for the X values transformed on the super level

        """

        if isinstance(X, DataFrame): 
            X_values = X.to_numpy()
        else:
            X_values = X    

        if latent_variables == None : 
            latent_variables = sum(self.latent_variables)
        else:
            latent_variables = sum(latent_variables)
        
        if isnan(sum(X_values)):
            
            result = zeros((X_values.shape[0], latent_variables))
            X_nan = isnan(X_values)
            variables_missing_mask = unique(X_nan, axis = 0)

            for row_mask in variables_missing_mask:
                
                rows_indexes = where((X_nan == row_mask).all(axis = 1))                
                
                if sum(row_mask) == 0: 

                    result[rows_indexes, :] = X[rows_indexes, :] @ self.x_weights_star[:latent_variables, :].T 
                
                else:
                    
                    result[rows_indexes, :] = scores_with_missing_values(self.omega, self.x_weights_star[:, ~row_mask], X[rows_indexes[0][:, None], ~row_mask], 
                                                                            LVs = latent_variables, method = self.missing_values_method)
                    
        else : result = X_values @ self.x_weights_star[:latent_variables, :].T

        # TO DO : check if X makes sense with latent variables
        return result

    def transform_inv(self, scores, latent_variables = None):

        """
        Transforms the superlevel scores matrix to the original X.

        Parameters
        ----------
        scores : array_like
            Matrix with all the scores to be used to rebuild X
        latent_variables : int, optional
            number of latent variables to be used. 

        Returns
        -------
        result : array_like 
            matrix of rebuilt X from scores
        """
                  
        if latent_variables == None : 
            latent_variables = sum(self.latent_variables)
        else:
            latent_variables = sum(latent_variables)

        result = scores @ self.p[:latent_variables, :] 
        
        return result
                     
    def predict( self, X, latent_variables = None): 

        """
        Predicts Y values using X array.

        Parameters
        ----------
        X : array_like
            Samples Matrix
        latent_variables : [int], optional
            number of latent variables to be used. 

        Returns
        -------
        preds : array_like 
            returns predictions
        """

        if isinstance( X, DataFrame ) : X = X.to_numpy()        
        if latent_variables == None : 
            latent_variables = sum(self.latent_variables)
        else:
            latent_variables = sum(latent_variables)

        preds = self.transform(X, latent_variables = latent_variables) @ self.c_loadings[ :latent_variables, :]
        
        return preds

        if latent_variables == None : latent_variables = self.latent_variables
     
        Y_hat = self.transform(X, latent_variables = latent_variables) @ self.c_loadings[:, :sum(latent_variables)].T

        return Y_hat

    def score(self, X, Y, latent_variables = None): 

        """
        Return the coefficient of determination R^2 of the prediction.

        R² is defined as 1 - Variance(Error) / Variance(Y) with Error = Y - predictions(X)

        Parameters
        ----------
        X : array_like
            Matrix with all the X to be used
        Y : array_like
            Matrix with all the Y ground truth values
        latent_variables : [int], optional
            number of latent variables to be used. 

        Returns
        -------
        score : float 
            returns calculated r².
        """

        if latent_variables == None : latent_variables = self.latent_variables
        if isinstance(Y, DataFrame) or isinstance(Y, Series): 
            Y_values = array(Y.to_numpy(), ndmin = 2).T
        else: 
            Y_values = Y

        Y_hat = self.predict(X, latent_variables = latent_variables)
        F = Y_values - Y_hat
        score = 1 - nanvar(F) / nanvar(Y_values)

        return score
    
    def _MLSMBPLS_1LV(self, X, third_level_pairs, second_level_pairs, Y, missing_values):
        
        num_blocks3l = len(third_level_pairs)
        conv = 1
        loop = 0 
        p_3rd = zeros((1, third_level_pairs[-1][-1] - third_level_pairs[0][0]))
        wb_2nd = zeros((1, second_level_pairs[-1][-1] - second_level_pairs[0][0] + num_blocks3l))
        p_2nd = zeros((1, second_level_pairs[-1][-1]  - second_level_pairs[0][0] + num_blocks3l))

        u = array(Y[:, 0], ndmin = 2).T

        while conv > self.tol and loop < self.loop_limit:
            
            wb_3rd, wb_2nd[0, 0:num_blocks3l], t3_final, superT = self._3rdlevel_part(X, third_level_pairs, u, 0)

            start = num_blocks3l
            T1l = superT.copy()
            for i, pair in enumerate(second_level_pairs): #consequence of the second level stuff
                wbcorr, T_scores = self._correlated_part(superT, X[:, pair[0]:pair[1]], u, 0)
                #correlated part
                end = start + pair[1] - pair[0]
                wb_2nd[0, start:end] = wbcorr
                start = end
                T1l = concatenate([T1l, T_scores], axis = 1)
            
            #superlevel stuff 
            # calculating the superlevel weights and scores
            wt, tT = self._superlevel_part(T1l, u)

            # calculating the y side loadings and scores
            q, u_new = self._y_part(tT, Y, 0)

            conv = norm(u - u_new)
            u = u_new
            loop +=1

        
        for i, pair in enumerate(third_level_pairs): #third level
            p_3rd[0, pair[0]:pair[1]] = self._p_loadings(X[:, pair[0]:pair[1]], t3_final[:,i], 0)
        
        p_2nd[0, 0:num_blocks3l] = self._p_loadings(t3_final, T1l[:,0], 0)

        start = num_blocks3l
        for i, pair in enumerate(second_level_pairs):
            end = start + pair[1] - pair[0]
            p_2nd[0, start:end] = self._p_loadings(X[:, pair[0]:pair[1]], T1l[:,i+1], 0)
            start = end


        result_dict = {'w3l':wb_3rd, #weights 3rd level
                       'w2l':wb_2nd, #weights 2rd level
                       'w1l':wt.T, #weights 1rd level
                       'p3l':p_3rd, #p loadings 3rd level
                       'p2l':p_2nd, #p loadings 2rd level
                       'p1l':self._p_loadings(X, tT, 0), #p loadings 1rd level
                       't3l':t3_final, #scores 3rd level
                       't2l':superT, #scores 2nd level
                       't1l':tT, #scores 1st level
                       'u':u,   #y training scores
                       'c':q.T}   #c loadings

        return result_dict

    def _SMBPLS_1LV(self, X, block_coord_pairs, Y, missing_values, block):
        
        conv = 1
        loops = 0

        y_scores1 = nan_to_num(array(Y[ :,0 ], ndmin = 2).T) #handling possible NaNs that come from faulty datasets - Step 1
        superlevel_scores1 = nan_to_num(y_scores1) #initializing superlevel scores vector

        while (conv > self.tol and loops < self.loop_limit) :

            first_block_done = False
            for missing, (start, end) in zip(missing_values, block_coord_pairs) :

                # calculates the block weights and scores for the uncorrelated part 
                if not first_block_done: 
                    block_weights, T_scores = self._uncorrelated_part(X[:, start:end], y_scores1, missing)
                    first_block_done = True
                    continue

                # calculating the block weights and scores for the blocks following the correlated parts of the following blocks
                corr_block_weights, corr_T_scores =  self._correlated_part(array(T_scores[:, 0], ndmin = 2).T, X[:, start:end], y_scores1, missing)
                block_weights = append(block_weights, corr_block_weights, axis = 1)
                T_scores = append(T_scores, corr_T_scores, axis = 1)
            
            # calculating the superlevel weights and scores
            superlevel_weights, superlevel_scores = self._superlevel_part(T_scores, y_scores1)

            # calculating the y side loadings and scores
            c_loadings, y_scores = self._y_part(superlevel_scores, Y, missing_values[-1])

            conv = norm(superlevel_scores1 - superlevel_scores) / norm(superlevel_scores1)
            superlevel_scores1 = superlevel_scores
            y_scores1 = y_scores

        # Calculating the p_loadings that connect from the raw X values to the superlevel scores
        start, end = block_coord_pairs[0][0], block_coord_pairs[-1][-1]
        superlevel_p_loadings = self._p_loadings(X[:, start:end], superlevel_scores, True in missing_values[:-1]).T
        x_weights = self._p_loadings(X[:, start:end], y_scores, True in missing_values).T
        
        x_weights = x_weights / norm(x_weights)
        y_scores_test = X[:, start:end] @ x_weights

        for missing, scores, (start, end) in zip(missing_values, T_scores.T, block_coord_pairs) :
            
            if start == block_coord_pairs[0][0] : 
                block_p_loadings = self._p_loadings(X[:, start:end], array(scores, ndmin = 2).T, missing)
                if start > 0:
                    block_p_loadings = append(zeros((1, start)), block_p_loadings, axis = 1 )
                    superlevel_p_loadings = append(zeros((start, 1)), superlevel_p_loadings, axis = 0 )
                    x_weights = append(zeros((start, 1)), x_weights, axis = 0 )
                    superlevel_weights = append(zeros((block, 1)), superlevel_weights, axis = 0 )
                    block_weights = append(zeros((1, start)), block_weights, axis = 1 )

            else : 
                block_p_loadings = append(block_p_loadings, self._p_loadings(X[:, start:end], array(scores, ndmin = 2).T, missing), axis = 1)

        result_dict = {'w2l':block_weights,
                       'w1l':x_weights.T,
                       'p2l':block_p_loadings,
                       't2l':T_scores,
                       'w1l':superlevel_weights.T,
                       'p1l':superlevel_p_loadings.T,
                       't1l':superlevel_scores,
                       'u':y_scores,
                       'c':c_loadings.T}

        return result_dict
    
    def _3rdlevel_part(self, X, third_level_pairs, y_scores, missing_value):
        
        wb_3rd = zeros((1, third_level_pairs[-1][-1] - third_level_pairs[0][0]))
        for i, pair in enumerate(third_level_pairs): #third level
                
            wb = array(X[:, pair[0]:pair[1]].T @ y_scores / (y_scores.T @ y_scores), ndmin = 2).T
            wb /= norm(wb)
            wb_3rd[0, pair[0]:pair[1]] = wb
            t = X[:, pair[0]:pair[1]] @ wb.T
            if i < 1: 
                t3_final = t
            else:
                t3_final = concatenate([t3_final, t], axis = 1)

        #consequence of the third level
        wb12 = array(t3_final.T @ y_scores / (y_scores.T @ y_scores), ndmin = 2).T
        wb12 /= norm(wb12)
        superT = t3_final @ wb12.T


        return wb_3rd, wb12, t3_final, superT

    def _uncorrelated_part(self, X, y_scores, missing_value):

        if missing_value:
            block_weights = array(nansum((X * y_scores), axis = 0), ndmin = 2) 
            block_weights = block_weights / nansum((array(~isnan(sum(X, axis = 1)), ndmin = 2).T * y_scores) ** 2, axis = 0)
            block_weights = block_weights / norm(block_weights)
            T_scores = array(nansum(X * block_weights, axis = 1) / nansum(((~isnan(X) * block_weights) ** 2), axis = 1), ndmin = 2).T
        
        else:
            block_weights = X.T @ y_scores / (y_scores.T @ y_scores) # calculating Xb block weights (as step 2.1 in L-G's 2018 paper)
            block_weights = (block_weights / norm(block_weights)).T
            T_scores = X @ block_weights.T
        
        return block_weights, T_scores
    
    def _correlated_part(self, scores, X, y_scores, missing_value):
        
        X_corr_coeffs = (scores / (scores.T @ scores)) @ scores.T
        
        if missing_value :     
            X_corr = X_corr_coeffs @ nan_to_num(X)  # Attention on this part
            place(X_corr, isnan(X), nan) # Keeping the NaN value as an NaN
            block_weights = array(nansum((X_corr * y_scores), axis = 0), ndmin = 2) 
            block_weights = block_weights / nansum((array(~isnan(sum(X, axis = 1)), ndmin = 2).T * y_scores) ** 2, axis = 0)
            block_weights = block_weights / norm(block_weights) #step 2.6
            T_scores =  array(nansum(X_corr * block_weights.T, axis = 1), ndmin = 2).T # step 2.7
        else :    
            X_corr = X_corr_coeffs @ X # finishing step 2.4 for no missing data
            block_weights = X_corr.T @ y_scores / (y_scores.T @ y_scores) # step 2.5
            block_weights = (block_weights / norm(block_weights)).T #step 2.6
            T_scores =  X_corr @ block_weights.T # step 2.7                
        
        return block_weights, T_scores
    
    def _superlevel_part(self, T_scores, y_scores):
        
        superlevel_weights = T_scores.T @ y_scores / (y_scores.T @ y_scores)  #step 2.9
        superlevel_weights = superlevel_weights / norm(superlevel_weights) #step 2.10
        superlevel_scores = T_scores @ superlevel_weights / (superlevel_weights.T @ superlevel_weights) #step 2.11

        return superlevel_weights, superlevel_scores

    def _y_part(self, superlevel_scores, Y, missing_value_Y):

        if missing_value_Y :
            c_loadings = nansum((Y.T * superlevel_scores).T, axis = 0) 
            c_loadings = c_loadings / nansum(((isnan(Y).T * superlevel_scores) ** 2).T, axis = 0)
        else : c_loadings = Y.T @ superlevel_scores / (superlevel_scores.T @ superlevel_scores) # step 2.12
        
        y_scores = Y @ c_loadings / (c_loadings.T @ c_loadings) # step 2.13

        return c_loadings, y_scores

    def _p_loadings(self, X, scores, missing_value):

        if missing_value:
            p_loadings = nansum(X * scores, axis = 0) 
            p_loadings = array(p_loadings / nansum(scores ** 2), ndmin = 2) #Step 3.1
            
        else: p_loadings = array(X.T @ scores / (scores.T @ scores), ndmin = 2) #Step 3.1
        if p_loadings.shape[0]> p_loadings.shape[1] : p_loadings = p_loadings.T
        return p_loadings

    def _cross_validate(self, X_orig, block_divs, Y_orig, LVs, random_state):

        
        cv_folds = KFold(n_splits = self.cv_splits_number, shuffle = True, random_state = random_state)
        q2 = []

        for train_index, test_index in cv_folds.split(X_orig):
            cv_model = SMB_PLS(tol = self.tol)
            cv_model.fit(X_orig[train_index], block_divs, Y_orig[train_index], latent_variables = LVs, int_call = True)
            q2.append(cv_model.score(X_orig[test_index], Y_orig[test_index]))

        q2 = mean(q2)

        return q2

    def _VIPs_calc( self, X, Y ): # code for calculation of VIPs

        """
        Calculates the VIP scores for all the variables for the prediction
        """
        
        SSY = sum((Y - nanmean(Y, axis = 0)) ** 2)
        for i in range(1, self.x_weights_star.shape[1] + 1):
            pred = self.predict(X, latent_variables = i)
            res = Y - pred
            SSY.loc[i] = sum(((res - res.mean(axis = 0)) ** 2))
            
        SSYdiff = SSY.iloc[:-1] - SSY.iloc[1:]
        VIPs = (((self.x_weights ** 2) @ SSYdiff.values) * self.weights.shape[1] / (SSY[0] - SSY[-1]) ** 1 / 2)
       
        return VIPs


import pandas as pd
from numpy import array, max
import pytest
import sys
import os
sys.path.append(".")
from trendfitter.models import MB_PCA

TESTDATA1_FILENAME = os.path.join(os.path.dirname(__file__), 'pca_test_dataset_complete.csv')
TESTDATA2_FILENAME = os.path.join(os.path.dirname(__file__), 'pca_test_dataset_incomplete.csv')

class TestMBPCA(object):

    def test_fit_fulldataset(self):
        
        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col=0)
        test_data = (test_data - test_data.mean()) / test_data.std()
        test_model = MB_PCA()
        test_model.fit(test_data, [2, 5])

        b_loadings = array([[-0.45753343,  0.4787455,  -0.53238767,  0.50447688, -0.15340262],
                            [-0.37043885,  0.35674997,  0.19766103, -0.22123992,  0.8046661 ]])
        assert test_model.block_loadings[:2, :] == pytest.approx(b_loadings), 'Block loadings are not as expected'

        splevel_loadings = array([[ 0.43853409,  0.56146591,  0.3402806,  -0.3402806, ],
                                  [ 0.3402806,  -0.3402806,   0.26449548,  0.73550452,]])
        assert test_model.superlevel_loadings == pytest.approx(splevel_loadings), 'superlevel loadings are not as expected'

        splevel_scores = array([[ 0.57628739,  0.80692377,  0.44244208, -1.06184817],
                                [-0.90296265, -1.89151428, -0.68209329,  1.03581792],
                                [ 0.33565122, -0.57320763,  0.26868383,  0.59277439],
                                [ 0.38422145,  1.55511912,  0.29561357, -1.43735021],
                                [ 0.71067313,  0.55726393,  0.54628867,  0.11717591]])

        assert test_model.superlevel_training_scores[:5] == pytest.approx(splevel_scores), 'superlevel scores are not as expected'

 
 
    def test_predict_b_fulldata(self):
        
        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col=0)
        test_data = (test_data - test_data.mean()) / test_data.std()
        test_model = MB_PCA()
        test_model.fit(test_data, [2, 5])

        expected = array([[-0.42756848,  0.43373619],
                          [ 0.66580945, -0.67562606],
                          [-0.25310258,  0.25654446],
                          [-0.28530091,  0.28940442],
                          [-0.52752326,  0.53512003]])

        assert expected == pytest.approx(test_model.predict_b(test_data.iloc[:5, :2], 0)), 'Prediction of block X values unsuccessful'
    
    
    def test_predict_b_missingdata(self):
        
        test_data = pd.read_csv(TESTDATA2_FILENAME, index_col = 0, delimiter = ';')
        test_data = (test_data - test_data.mean()) / test_data.std()
        test_model = MB_PCA()
        test_model.fit(test_data, [2, 5])

        expected = array([[-0.40119088,  0.39775356],
                          [ 0.62157011, -0.61701447],
                          [-0.240606,    0.23778596],
                          [-0.26787863,  0.26554006],
                          [-0.49517005,  0.49088111]])

        assert expected == pytest.approx(test_model.predict_b(test_data.iloc[:5, :2], 0)), 'Prediction of X values with missing data unsuccessful'
    
    
    def test_transform_b_fulldata(self):

        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col=0)
        test_data = (test_data - test_data.mean()) / test_data.std()
        test_model = MB_PCA()
        test_model.fit(test_data, [2, 5])
        
        expected = array([[ 0.57628739,  0.44244208],
                          [-0.90296265, -0.68209329],
                          [ 0.33565122,  0.26868383],
                          [ 0.38422145,  0.29561357],
                          [ 0.71067313,  0.54628867]])

        assert expected == pytest.approx(test_model.transform_b(test_data.iloc[:5, :2], 0)), "Transformation to block scores is malfunctioning "

    def test_transform_b_missingdata(self):

        test_data = pd.read_csv(TESTDATA2_FILENAME, index_col = 0, delimiter = ';')
        test_data = (test_data - test_data.mean()) / test_data.std()
        test_model = MB_PCA()
        test_model.fit(test_data, [2, 5])

        expected = array([[ 0.58309938,  0.38480606],
                          [-0.91347032, -0.58249535],
                          [ 0.33978214,  0.24426879],
                          [ 0.38877249,  0.25771076],
                          [ 0.71908363,  0.47577261]])

        assert expected == pytest.approx(test_model.transform_b(test_data.iloc[:5, :2], 0))

    
    def test_score_b(self):
        
        test_data = pd.read_csv(TESTDATA1_FILENAME, index_col = 0)
        test_data = (test_data - test_data.mean()) / test_data.std()
        test_model = MB_PCA()
        test_model.fit(test_data, [2, 5])
        
        expected = 0.7977328443400561
        
        assert expected == pytest.approx(test_model.score_b(test_data.iloc[:, :2], 0)), "block r² Score calculation malfunctioning"
        
